from __future__ import annotations

import asyncio
import heapq
import logging
import os
import pickle
import random
import re
import secrets
import time
import uuid
from abc import ABC, abstractmethod
from base64 import b64encode
from collections import Counter, OrderedDict
from hashlib import sha1
from itertools import chain
from operator import itemgetter
from typing import Any, Coroutine, NamedTuple

import msgpack

log = logging.getLogger(__name__)


class Value(NamedTuple):
    value: Any
    ttl: float | None
    ts: float

    def outdated(self):
        return self.ttl is not None and self.ttl < time.time() - self.ts

    def __repr__(self):
        fmt = "%a, %d %b %Y %H:%M:%S"
        mt = time.strftime(fmt, time.localtime(self.ts))
        return f"{self.value!r}, ttl={self.ttl}, last modified at {mt}"


class IStorage(ABC):
    """
    Local storage for this node.
    IStorage implementations of get must return the same type as put in by set
    """

    @abstractmethod
    def __setitem__(self, key: bytes, value):
        """
        Set a key to the given value.
        """

    @abstractmethod
    def __getitem__(self, key: bytes):
        """
        Get the given key.  If item doesn't exist, return None.
        """

    @abstractmethod
    def get(self, key: bytes, default=None):
        """
        Get given key.  If not found, return default.
        """

    @abstractmethod
    def set(self, key: bytes, value, ttl: float = 0):
        """
        Set a key to the given value.
        """

    @abstractmethod
    def iter_older_than(self, seconds_old: float):
        """
        Return the an iterator over (key, value) tuples for items older
        than the given secondsOld.
        """

    @abstractmethod
    def __iter__(self):
        """
        Get the iterator for this storage, should yield tuple of (key, value)
        """


class ForgetfulStorage(IStorage):

    def __init__(self, ttl: float = 604800):
        self._storage: dict[bytes, Value] = {}
        self.ttl: float = ttl

    def get(self, key: bytes):
        if key not in self._storage:
            return None
        value = self._storage[key]
        if value.outdated():
            del self._storage[key]
            return None
        else:
            return value.value

    def set(self, key: bytes, value, ttl: float | None = None):
        if ttl is None:
            ttl = self.ttl
        self._storage[key] = Value(value, ttl, time.time())

    def __iter(self):
        for key, value in list(self._storage.items()):
            if value.outdated():
                del self._storage[key]
                continue
            yield key, value

    def cull(self):
        for key, value in self.__iter():
            if value.outdated():
                del self._storage[key]
                continue

    def __repr__(self):
        self.cull()
        return repr(self._storage)

    def __contains__(self, key: bytes):
        if key in self._storage:
            value = self._storage[key]
            if value.outdated():
                del self._storage[key]
                return False
            else:
                return True
        else:
            return False

    def __setitem__(self, key: bytes, value):
        self.set(key, value, ttl=self.ttl)

    def __getitem__(self, key: bytes):
        return self.get(key)

    def __iter__(self):
        for key, value in self.__iter():
            yield key, value.value

    def expire(self, key: bytes, ttl: float):
        if key in self._storage:
            value = self._storage[key]
            self._storage[key] = Value(value.value, ttl, time.time())

    def exists(self, key: bytes):
        return self.__contains__(key)

    def keys(self, pattern="*"):
        return [key for key in self.scan_iter(pattern)]

    def scan_iter(self, pattern="*"):
        prog = re.compile(pattern.replace("*", ".*"))
        for key, _ in self.__iter():
            if prog.match(key):
                yield key

    def iter_older_than(self, t: float):
        for key, value in self.__iter():
            if time.time() - value.ts >= t:
                yield key, value.value


async def gather_dict(dic: dict[bytes, Coroutine]):
    cors = list(dic.values())
    results = await asyncio.gather(*cors)
    return dict(zip(dic.keys(), results))


def digest(string: str | bytes) -> bytes:
    if not isinstance(string, bytes):
        string = str(string).encode('utf8')
    return sha1(string).digest()


def shared_prefix(args):
    """
    Find the shared prefix between the strings.
    For instance:
        sharedPrefix(['blahblah', 'blahwhat'])
    returns 'blah'.
    """
    i = 0
    while i < min(map(len, args)):
        if len(set(map(itemgetter(i), args))) != 1:
            break
        i += 1
    return args[0][:i]


def bytes_to_bit_string(bites):
    bits = [bin(bite)[2:].rjust(8, '0') for bite in bites]
    return "".join(bits)


class Node:
    """
    Simple object to encapsulate the concept of a Node (minimally an ID, but
    also possibly an IP and port if this represents a node on the network).
    This class should generally not be instantiated directly, as it is a low
    level construct mostly used by the router.
    """
    __slots__ = ('id', 'ip', 'port', 'long_id')

    def __init__(self, node_id: bytes, ip=None, port=None):
        """
        Create a Node instance.
        Args:
            node_id (int): A value between 0 and 2^160
            ip (string): Optional IP address where this Node lives
            port (int): Optional port for this Node (set when IP is set)
        """
        self.id = node_id  # pylint: disable=invalid-name
        self.ip = ip  # pylint: disable=invalid-name
        self.port = port
        self.long_id = int(node_id.hex(), 16)

    def same_home_as(self, node):
        return self.ip == node.ip and self.port == node.port

    def distance_to(self, node):
        """
        Get the distance between this node and another.
        """
        return self.long_id ^ node.long_id

    def __iter__(self):
        """
        Enables use of Node as a tuple - i.e., tuple(node) works.
        """
        return iter([self.id, self.ip, self.port])

    def __repr__(self):
        return repr([self.long_id, self.ip, self.port])

    def __str__(self):
        return "%s:%s" % (self.ip, str(self.port))


class NodeHeap:
    """
    A heap of nodes ordered by distance to a given node.
    """

    def __init__(self, node: Node, maxsize: int):
        """
        Constructor.
        @param node: The node to measure all distnaces from.
        @param maxsize: The maximum size that this heap can grow to.
        """
        self.node = node
        self.heap = []
        self.contacted = set()
        self.maxsize = maxsize

    def remove(self, peers):
        """
        Remove a list of peer ids from this heap.  Note that while this
        heap retains a constant visible size (based on the iterator), it's
        actual size may be quite a bit larger than what's exposed.  Therefore,
        removal of nodes may not change the visible size as previously added
        nodes suddenly become visible.
        """
        peers = set(peers)
        if not peers:
            return
        nheap = []
        for distance, node in self.heap:
            if node.id not in peers:
                heapq.heappush(nheap, (distance, node))
        self.heap = nheap

    def get_node(self, node_id):
        for _, node in self.heap:
            if node.id == node_id:
                return node
        return None

    def have_contacted_all(self):
        return len(self.get_uncontacted()) == 0

    def get_ids(self):
        return [n.id for n in self]

    def mark_contacted(self, node):
        self.contacted.add(node.id)

    def popleft(self):
        return heapq.heappop(self.heap)[1] if self else None

    def push(self, nodes):
        """
        Push nodes onto heap.
        @param nodes: This can be a single item or a C{list}.
        """
        if not isinstance(nodes, list):
            nodes = [nodes]

        for node in nodes:
            if node not in self:
                distance = self.node.distance_to(node)
                heapq.heappush(self.heap, (distance, node))

    def __len__(self):
        return min(len(self.heap), self.maxsize)

    def __iter__(self):
        nodes = heapq.nsmallest(self.maxsize, self.heap)
        return iter(map(itemgetter(1), nodes))

    def __contains__(self, node):
        for _, other in self.heap:
            if node.id == other.id:
                return True
        return False

    def get_uncontacted(self):
        return [n for n in self if n.id not in self.contacted]


class KBucket:

    def __init__(self, rangeLower, rangeUpper, ksize, replacementNodeFactor=5):
        self.range = (rangeLower, rangeUpper)
        self.nodes = OrderedDict()
        self.replacement_nodes = OrderedDict()
        self.touch_last_updated()
        self.ksize = ksize
        self.max_replacement_nodes = self.ksize * replacementNodeFactor

    def touch_last_updated(self):
        self.last_updated = time.monotonic()

    def get_nodes(self):
        return list(self.nodes.values())

    def split(self):
        midpoint = (self.range[0] + self.range[1]) // 2
        one = KBucket(self.range[0], midpoint, self.ksize)
        two = KBucket(midpoint + 1, self.range[1], self.ksize)
        nodes = chain(self.nodes.values(), self.replacement_nodes.values())
        for node in nodes:
            bucket = one if node.long_id <= midpoint else two
            bucket.add_node(node)

        return (one, two)

    def remove_node(self, node):
        if node.id in self.replacement_nodes:
            del self.replacement_nodes[node.id]

        if node.id in self.nodes:
            del self.nodes[node.id]

            if self.replacement_nodes:
                newnode_id, newnode = self.replacement_nodes.popitem()
                self.nodes[newnode_id] = newnode

    def has_in_range(self, node):
        return self.range[0] <= node.long_id <= self.range[1]

    def is_new_node(self, node):
        return node.id not in self.nodes

    def add_node(self, node):
        """
        Add a C{Node} to the C{KBucket}.  Return True if successful,
        False if the bucket is full.
        If the bucket is full, keep track of node in a replacement list,
        per section 4.1 of the paper.
        """
        if node.id in self.nodes:
            del self.nodes[node.id]
            self.nodes[node.id] = node
        elif len(self) < self.ksize:
            self.nodes[node.id] = node
        else:
            if node.id in self.replacement_nodes:
                del self.replacement_nodes[node.id]
            self.replacement_nodes[node.id] = node
            while len(self.replacement_nodes) > self.max_replacement_nodes:
                self.replacement_nodes.popitem(last=False)
            return False
        return True

    def depth(self):
        vals = self.nodes.values()
        sprefix = shared_prefix([bytes_to_bit_string(n.id) for n in vals])
        return len(sprefix)

    def head(self):
        return list(self.nodes.values())[0]

    def __getitem__(self, node_id):
        return self.nodes.get(node_id, None)

    def __len__(self):
        return len(self.nodes)


class TableTraverser:

    def __init__(self, table, startNode):
        index = table.get_bucket_for(startNode)
        table.buckets[index].touch_last_updated()
        self.current_nodes = table.buckets[index].get_nodes()
        self.left_buckets = table.buckets[:index]
        self.right_buckets = table.buckets[(index + 1):]
        self.left = True

    def __iter__(self):
        return self

    def __next__(self):
        """
        Pop an item from the left subtree, then right, then left, etc.
        """
        if self.current_nodes:
            return self.current_nodes.pop()

        if self.left and self.left_buckets:
            self.current_nodes = self.left_buckets.pop().get_nodes()
            self.left = False
            return next(self)

        if self.right_buckets:
            self.current_nodes = self.right_buckets.pop(0).get_nodes()
            self.left = True
            return next(self)

        raise StopIteration


class RoutingTable:

    def __init__(self, protocol, ksize, node):
        """
        @param node: The node that represents this server.  It won't
        be added to the routing table, but will be needed later to
        determine which buckets to split or not.
        """
        self.node = node
        self.protocol = protocol
        self.ksize = ksize
        self.flush()

    def flush(self):
        self.buckets: list[KBucket] = [KBucket(0, 2**160, self.ksize)]

    def split_bucket(self, index: int):
        one, two = self.buckets[index].split()
        self.buckets[index] = one
        self.buckets.insert(index + 1, two)

    def lonely_buckets(self):
        """
        Get all of the buckets that haven't been updated in over
        an hour.
        """
        hrago = time.monotonic() - 3600
        return [b for b in self.buckets if b.last_updated < hrago]

    def remove_contact(self, node):
        index = self.get_bucket_for(node)
        self.buckets[index].remove_node(node)

    def is_new_node(self, node):
        index = self.get_bucket_for(node)
        return self.buckets[index].is_new_node(node)

    def add_contact(self, node):
        index = self.get_bucket_for(node)
        bucket = self.buckets[index]

        # this will succeed unless the bucket is full
        if bucket.add_node(node):
            return

        # Per section 4.2 of paper, split if the bucket has the node
        # in its range or if the depth is not congruent to 0 mod 5
        if bucket.has_in_range(self.node) or bucket.depth() % 5 != 0:
            self.split_bucket(index)
            self.add_contact(node)
        else:
            asyncio.ensure_future(self.protocol.call_ping(bucket.head()))

    def get_bucket_for(self, node):
        """
        Get the index of the bucket that the given node would fall into.
        """
        for index, bucket in enumerate(self.buckets):
            if node.long_id < bucket.range[1]:
                return index
        # we should never be here, but make linter happy
        return None

    def find_neighbors(self, node, k=None, exclude=None):
        k = k or self.ksize
        nodes = []
        for neighbor in TableTraverser(self, node):
            notexcluded = exclude is None or not neighbor.same_home_as(exclude)
            if neighbor.id != node.id and notexcluded:
                heapq.heappush(nodes, (node.distance_to(neighbor), neighbor))
            if len(nodes) == k:
                break

        return list(map(itemgetter(1), heapq.nsmallest(k, nodes)))


class SpiderCrawl:
    """
    Crawl the network and look for given 160-bit keys.
    """

    def __init__(self, protocol: KademliaProtocol, node, peers, ksize, alpha):
        """
        Create a new C{SpiderCrawl}er.
        Args:
            protocol: A :class:`~kademlia.protocol.KademliaProtocol` instance.
            node: A :class:`~kademlia.node.Node` representing the key we're
                  looking for
            peers: A list of :class:`~kademlia.node.Node` instances that
                   provide the entry point for the network
            ksize: The value for k based on the paper
            alpha: The value for alpha based on the paper
        """
        self.protocol = protocol
        self.ksize = ksize
        self.alpha = alpha
        self.node = node
        self.nearest = NodeHeap(self.node, self.ksize)
        self.last_ids_crawled = []
        log.info("creating spider with peers: %s", peers)
        self.nearest.push(peers)

    async def _find(self, rpcmethod):
        """
        Get either a value or list of nodes.
        Args:
            rpcmethod: The protocol's callfindValue or call_find_node.
        The process:
          1. calls find_* to current ALPHA nearest not already queried nodes,
             adding results to current nearest list of k nodes.
          2. current nearest list needs to keep track of who has been queried
             already sort by nearest, keep KSIZE
          3. if list is same as last time, next call should be to everyone not
             yet queried
          4. repeat, unless nearest list has all been queried, then ur done
        """
        log.info("crawling network with nearest: %s", str(tuple(self.nearest)))
        count = self.alpha
        if self.nearest.get_ids() == self.last_ids_crawled:
            count = len(self.nearest)
        self.last_ids_crawled = self.nearest.get_ids()

        dicts = {}
        for peer in self.nearest.get_uncontacted()[:count]:
            dicts[peer.id] = rpcmethod(peer, self.node)
            self.nearest.mark_contacted(peer)
        found = await gather_dict(dicts)
        return await self._nodes_found(found)

    async def _nodes_found(self, responses):
        raise NotImplementedError


class ValueSpiderCrawl(SpiderCrawl):

    def __init__(self, protocol: KademliaProtocol, node, peers, ksize, alpha):
        SpiderCrawl.__init__(self, protocol, node, peers, ksize, alpha)
        # keep track of the single nearest node without value - per
        # section 2.3 so we can set the key there if found
        self.nearest_without_value = NodeHeap(self.node, 1)

    async def find(self):
        """
        Find either the closest nodes or the value requested.
        """
        return await self._find(self.protocol.call_find_value)

    async def _nodes_found(self, responses):
        """
        Handle the result of an iteration in _find.
        """
        toremove = []
        found_values = []
        for peerid, response in responses.items():
            response = RPCFindResponse(response)
            if not response.happened():
                toremove.append(peerid)
            elif response.has_value():
                found_values.append(response.get_value())
            else:
                peer = self.nearest.get_node(peerid)
                self.nearest_without_value.push(peer)
                self.nearest.push(response.get_node_list())
        self.nearest.remove(toremove)

        if found_values:
            return await self._handle_found_values(found_values)
        if self.nearest.have_contacted_all():
            # not found!
            return None
        return await self.find()

    async def _handle_found_values(self, values):
        """
        We got some values!  Exciting.  But let's make sure
        they're all the same or freak out a little bit.  Also,
        make sure we tell the nearest node that *didn't* have
        the value to store it.
        """
        value_counts = Counter(values)
        if len(value_counts) != 1:
            log.warning("Got multiple values for key %i: %s",
                        self.node.long_id, str(values))
        value = value_counts.most_common(1)[0][0]

        peer = self.nearest_without_value.popleft()
        if peer:
            await self.protocol.call_store(peer, self.node.id, value)
        return value


class NodeSpiderCrawl(SpiderCrawl):

    async def find(self):
        """
        Find the closest nodes.
        """
        return await self._find(self.protocol.call_find_node)

    async def _nodes_found(self, responses):
        """
        Handle the result of an iteration in _find.
        """
        toremove = []
        for peerid, response in responses.items():
            response = RPCFindResponse(response)
            if not response.happened():
                toremove.append(peerid)
            else:
                self.nearest.push(response.get_node_list())
        self.nearest.remove(toremove)

        if self.nearest.have_contacted_all():
            return list(self.nearest)
        return await self.find()


class RPCFindResponse:

    def __init__(self, response):
        """
        A wrapper for the result of a RPC find.
        Args:
            response: This will be a tuple of (<response received>, <value>)
                      where <value> will be a list of tuples if not found or
                      a dictionary of {'value': v} where v is the value desired
        """
        self.response = response

    def happened(self):
        """
        Did the other host actually respond?
        """
        return self.response[0]

    def has_value(self):
        return isinstance(self.response[1], dict)

    def get_value(self):
        return self.response[1]['value']

    def get_node_list(self):
        """
        Get the node list in the response.  If there's no value, this should
        be set.
        """
        nodelist = self.response[1] or []
        return [Node(*nodeple) for nodeple in nodelist]


class MalformedMessage(Exception):
    """
    Message does not contain what is expected.
    """


class RPCProtocol(asyncio.DatagramProtocol):

    def __init__(self, wait_timeout: float = 5):
        self._timeout = wait_timeout
        self._outstanding: dict[bytes, tuple[asyncio.Future,
                                             asyncio.TimerHandle]] = {}
        self.transport = None

    def connection_made(self, transport):
        self.transport = transport

    def datagram_received(self, datagram, address):
        log.debug("received datagram from %s", address)
        if len(datagram) < 22:
            log.warning("received datagram too small from %s,"
                        " ignoring", address)
            return

        msg_id = datagram[1:21]
        data = msgpack.unpackb(datagram[21:])

        if datagram[:1] == b'\x00':
            # schedule accepting request and returning the result
            self.handle_request(msg_id, data, address)
        elif datagram[:1] == b'\x01':
            self.handle_response(msg_id, data, address)
        else:
            # otherwise, don't know the format, don't do anything
            log.debug("Received unknown message from %s, ignoring", address)

    def handle_response(self, msg_id, data, address):
        msgargs = (b64encode(msg_id), address)
        if msg_id not in self._outstanding:
            log.warning("received unknown message %s "
                        "from %s; ignoring", *msgargs)
            return
        log.debug("received response %s for message "
                  "id %s from %s", data, *msgargs)
        future, timeout = self._outstanding[msg_id]
        timeout.cancel()
        future.set_result((True, data))
        del self._outstanding[msg_id]

    def handle_request(self, msg_id, data, address):
        if not isinstance(data, list) or len(data) != 2:
            raise MalformedMessage("Could not read packet: %s" % data)
        funcname, args = data

        try:
            func = getattr(self, f"on_{funcname}")
        except AttributeError:
            msgargs = (self.__class__.__name__, funcname)
            log.warning("%s has no callable method "
                        "on_%s; ignoring request", *msgargs)
            return

        try:
            response = func(address, *args)
        except Exception as e:
            log.exception(e)

        log.debug("sending response %s for msg id %s to %s", response,
                  b64encode(msg_id), address)
        txdata = b'\x01' + msg_id + msgpack.packb(response)
        self.transport.sendto(txdata, address)

    def rpc_call(self, name, address, *args):
        msg_id = sha1(os.urandom(32)).digest()
        data = msgpack.packb([name, args])
        if len(data) > 8192:
            raise MalformedMessage("Total length of function "
                                   "name and arguments cannot exceed 8K")
        txdata = b'\x00' + msg_id + data
        log.debug("calling remote function %s on %s (msgid %s)", name, address,
                  b64encode(msg_id))
        self.transport.sendto(txdata, address)

        loop = asyncio.get_event_loop()
        if hasattr(loop, 'create_future'):
            future = loop.create_future()
        else:
            future = asyncio.Future()
        timeout = loop.call_later(self._timeout, self.rpc_cancel, msg_id)
        self._outstanding[msg_id] = (future, timeout)
        return future

    def rpc_cancel(self, msg_id):
        args = (b64encode(msg_id), self._timeout)
        log.error("Did not receive reply for msg "
                  "id %s within %i seconds", *args)
        self._outstanding[msg_id][0].set_result((False, None))
        del self._outstanding[msg_id]


class KademliaProtocol(RPCProtocol):

    def __init__(self, source_node, storage, ksize, wait_timeout=5):
        super().__init__(wait_timeout)
        self.router = RoutingTable(self, ksize, source_node)
        self.storage = storage
        self.source_node = source_node

    def get_refresh_ids(self):
        """
        Get ids to search for to keep old buckets up to date.
        """
        ids = []
        for bucket in self.router.lonely_buckets():
            rid = random.randint(*bucket.range).to_bytes(20, byteorder='big')
            ids.append(rid)
        return ids

    def on_ping(self, sender, nodeid):
        source = Node(nodeid, sender[0], sender[1])
        self.welcome_if_new(source)
        return self.source_node.id

    def on_store(self, sender, nodeid, key, value):
        source = Node(nodeid, sender[0], sender[1])
        self.welcome_if_new(source)
        log.debug("got a store request from %s, storing '%s'='%s'", sender,
                  key.hex(), value)
        self.storage[key] = value
        return True

    def on_find_node(self, sender, nodeid, key):
        log.info("finding neighbors of %i in local table",
                 int(nodeid.hex(), 16))
        source = Node(nodeid, sender[0], sender[1])
        self.welcome_if_new(source)
        node = Node(key)
        neighbors = self.router.find_neighbors(node, exclude=source)
        return list(map(tuple, neighbors))

    def on_find_value(self, sender, nodeid, key):
        source = Node(nodeid, sender[0], sender[1])
        self.welcome_if_new(source)
        value = self.storage.get(key)
        if value is None:
            return self.on_find_node(sender, nodeid, key)
        return {'value': value}

    async def ping(self, address, source_node_id):
        return await self.rpc_call('ping', address, source_node_id)

    async def store(self, address, source_node_id, key, value):
        return await self.rpc_call('store', address, source_node_id, key,
                                   value)

    async def find_node(self, address, source_node_id, node_to_find_id):
        return await self.rpc_call('find_node', address, source_node_id,
                                   node_to_find_id)

    async def find_value(self, address, source_node_id, node_to_find_id):
        return await self.rpc_call('find_value', address, source_node_id,
                                   node_to_find_id)

    async def call_find_node(self, node_to_ask, node_to_find):
        address = (node_to_ask.ip, node_to_ask.port)
        result = await self.find_node(address, self.source_node.id,
                                      node_to_find.id)
        return self.handle_rpc_result(result, node_to_ask)

    async def call_find_value(self, node_to_ask, node_to_find):
        address = (node_to_ask.ip, node_to_ask.port)
        result = await self.find_value(address, self.source_node.id,
                                       node_to_find.id)
        return self.handle_rpc_result(result, node_to_ask)

    async def call_ping(self, node_to_ask):
        address = (node_to_ask.ip, node_to_ask.port)
        result = await self.ping(address, self.source_node.id)
        return self.handle_rpc_result(result, node_to_ask)

    async def call_store(self, node_to_ask, key, value):
        address = (node_to_ask.ip, node_to_ask.port)
        result = await self.store(address, self.source_node.id, key, value)
        return self.handle_rpc_result(result, node_to_ask)

    def welcome_if_new(self, node):
        """
        Given a new node, send it all the keys/values it should be storing,
        then add it to the routing table.
        @param node: A new node that just joined (or that we just found out
        about).
        Process:
        For each key in storage, get k closest nodes.  If newnode is closer
        than the furtherst in that list, and the node for this server
        is closer than the closest in that list, then store the key/value
        on the new node (per section 2.5 of the paper)
        """
        if not self.router.is_new_node(node):
            return

        log.info("never seen %s before, adding to router", node)
        for key, value in self.storage:
            keynode = Node(digest(key))
            neighbors = self.router.find_neighbors(keynode)
            if neighbors:
                last = neighbors[-1].distance_to(keynode)
                new_node_close = node.distance_to(keynode) < last
                first = neighbors[0].distance_to(keynode)
                this_closest = self.source_node.distance_to(keynode) < first
            if not neighbors or (new_node_close and this_closest):
                asyncio.ensure_future(self.call_store(node, key, value))
        self.router.add_contact(node)

    def handle_rpc_result(self, result, node):
        """
        If we get a response, add the node to the routing table.  If
        we get no response, make sure it's removed from the routing table.
        """
        if not result[0]:
            log.warning("no response from %s, removing from router", node)
            self.router.remove_contact(node)
            return result

        log.info("got successful response from %s", node)
        self.welcome_if_new(node)
        return result


class Server:
    """
    High level view of a node instance.  This is the object that should be
    created to start listening as an active node on the network.
    """

    protocol_class = KademliaProtocol

    def __init__(self, ksize=20, alpha=3, node_id=None, storage=None):
        """
        Create a server instance.  This will start listening on the given port.
        Args:
            ksize (int): The k parameter from the paper
            alpha (int): The alpha parameter from the paper
            node_id: The id for this node on the network.
            storage: An instance that implements the interface
                     :class:`~kademlia.storage.IStorage`
        """
        self.ksize = ksize
        self.alpha = alpha
        self.storage = storage or ForgetfulStorage()
        self.node = Node(node_id
                         or uuid.uuid1().bytes[4:] + secrets.token_bytes(8))
        self.transport = None
        self.protocol = None
        self.refresh_loop = None
        self.save_state_loop = None

    def stop(self):
        if self.transport is not None:
            self.transport.close()

        if self.refresh_loop:
            self.refresh_loop.cancel()

        if self.save_state_loop:
            self.save_state_loop.cancel()

    def _create_protocol(self):
        return self.protocol_class(self.node, self.storage, self.ksize)

    async def listen(self, port, interface='0.0.0.0', interval=300):
        """
        Start listening on the given port.
        Provide interface="::" to accept ipv6 address
        """
        loop = asyncio.get_event_loop()
        listen = loop.create_datagram_endpoint(self._create_protocol,
                                               local_addr=(interface, port))
        log.info("Node %i listening on %s:%i", self.node.long_id, interface,
                 port)
        self.transport, self.protocol = await listen
        # finally, schedule refreshing table
        self.refresh_table(interval)

    def refresh_table(self, interval=300):
        log.debug("Refreshing routing table")
        asyncio.ensure_future(self._refresh_table(interval))
        loop = asyncio.get_running_loop()
        self.refresh_loop = loop.call_later(interval, self.refresh_table,
                                            interval)

    async def _refresh_table(self, interval=300):
        """
        Refresh buckets that haven't had any lookups in the last hour
        (per section 2.3 of the paper).
        """
        results = []
        for node_id in self.protocol.get_refresh_ids():
            node = Node(node_id)
            nearest = self.protocol.router.find_neighbors(node, self.alpha)
            spider = NodeSpiderCrawl(self.protocol, node, nearest, self.ksize,
                                     self.alpha)
            results.append(spider.find())

        # do our crawling
        await asyncio.gather(*results)

        # now republish keys older than one hour
        for dkey, value in self.storage.iter_older_than(interval):
            await self.set_digest(dkey, value)

    def bootstrappable_neighbors(self):
        """
        Get a :class:`list` of (ip, port) :class:`tuple` pairs suitable for
        use as an argument to the bootstrap method.
        The server should have been bootstrapped
        already - this is just a utility for getting some neighbors and then
        storing them if this server is going down for a while.  When it comes
        back up, the list of nodes can be used to bootstrap.
        """
        neighbors = self.protocol.router.find_neighbors(self.node)
        return [tuple(n)[-2:] for n in neighbors]

    async def bootstrap(self, addrs):
        """
        Bootstrap the server by connecting to other known nodes in the network.
        Args:
            addrs: A `list` of (ip, port) `tuple` pairs.  Note that only IP
                   addresses are acceptable - hostnames will cause an error.
        """
        log.debug("Attempting to bootstrap node with %i initial contacts",
                  len(addrs))
        cos = list(map(self.bootstrap_node, addrs))
        gathered = await asyncio.gather(*cos)
        nodes = [node for node in gathered if node is not None]
        spider = NodeSpiderCrawl(self.protocol, self.node, nodes, self.ksize,
                                 self.alpha)
        return await spider.find()

    async def bootstrap_node(self, addr):
        result = await self.protocol.ping(addr, self.node.id)
        return Node(result[1], addr[0], addr[1]) if result[0] else None

    async def get(self, key):
        """
        Get a key if the network has it.
        Returns:
            :class:`None` if not found, the value otherwise.
        """
        log.info("Looking up key %s", key)
        dkey = digest(key)
        # if this node has it, return it
        if self.storage.get(dkey) is not None:
            return self.storage.get(dkey)
        node = Node(dkey)
        nearest = self.protocol.router.find_neighbors(node)
        if not nearest:
            log.warning("There are no known neighbors to get key %s", key)
            return None
        spider = ValueSpiderCrawl(self.protocol, node, nearest, self.ksize,
                                  self.alpha)
        return await spider.find()

    async def set(self, key, value):
        """
        Set the given string key to the given value in the network.
        """
        if not check_dht_value_type(value):
            raise TypeError(
                "Value must be of type int, float, bool, str, or bytes")
        log.info("setting '%s' = '%s' on network", key, value)
        dkey = digest(key)
        return await self.set_digest(dkey, value)

    async def set_digest(self, dkey, value):
        """
        Set the given SHA1 digest key (bytes) to the given value in the
        network.
        """
        node = Node(dkey)

        nearest = self.protocol.router.find_neighbors(node)
        if not nearest:
            log.warning("There are no known neighbors to set key %s",
                        dkey.hex())
            self.storage[dkey] = value
            return True

        spider = NodeSpiderCrawl(self.protocol, node, nearest, self.ksize,
                                 self.alpha)
        nodes = await spider.find()
        log.info("setting '%s' on %s", dkey.hex(), list(map(str, nodes)))

        # if this node is close too, then store here as well
        biggest = max([n.distance_to(node) for n in nodes])
        if self.node.distance_to(node) < biggest:
            self.storage[dkey] = value
        results = [self.protocol.call_store(n, dkey, value) for n in nodes]
        # return true only if at least one store call succeeded
        return any(await asyncio.gather(*results))

    def save_state(self, fname):
        """
        Save the state of this node (the alpha/ksize/id/immediate neighbors)
        to a cache file with the given fname.
        """
        log.info("Saving state to %s", fname)
        data = {
            'ksize': self.ksize,
            'alpha': self.alpha,
            'id': self.node.id,
            'neighbors': self.bootstrappable_neighbors(),
            'storage': self.storage
        }
        if not data['neighbors']:
            log.warning("No known neighbors, so not writing to cache.")
            # return
        with open(fname, 'wb') as file:
            pickle.dump(data, file)

    @classmethod
    async def load_state(cls, fname, port, interface='0.0.0.0', interval=300):
        """
        Load the state of this node (the alpha/ksize/id/immediate neighbors)
        from a cache file with the given fname and then bootstrap the node
        (using the given port/interface to start listening/bootstrapping).
        """
        log.info("Loading state from %s", fname)
        with open(fname, 'rb') as file:
            data = pickle.load(file)
        svr = cls(data['ksize'], data['alpha'], data['id'], data['storage'])
        await svr.listen(port, interface, interval)
        if data['neighbors']:
            await svr.bootstrap(data['neighbors'])
        return svr

    def save_state_regularly(self, fname, frequency=300):
        """
        Save the state of node with a given regularity to the given
        filename.
        Args:
            fname: File name to save retularly to
            frequency: Frequency in seconds that the state should be saved.
                        By default, 10 minutes.
        """
        self.save_state(fname)
        loop = asyncio.get_running_loop()
        self.save_state_loop = loop.call_later(frequency,
                                               self.save_state_regularly,
                                               fname, frequency)


def check_dht_value_type(value):
    """
    Checks to see if the type of the value is a valid type for
    placing in the dht.
    """
    return isinstance(value, (int, float, bool, str, bytes))



import dataclasses
import logging
import pickle
import re
import time
from concurrent.futures import ThreadPoolExecutor
from datetime import datetime
from pathlib import Path
from random import shuffle
from typing import Any, List, Optional, TypedDict

import numpy as np
import openai
import tenacity
import tiktoken
from IPython import get_ipython
from IPython.display import Markdown, display
from openai.error import (APIConnectionError, APIError, RateLimitError,
                          ServiceUnavailableError, Timeout)
from scipy import spatial

logger = logging.getLogger(__name__)


class Message(TypedDict):
    """OpenAI Message object containing a role and the message content"""

    role: str
    content: str


DEFAULT_SYSTEM_PROMPT = 'You are a helpful assistant. Respond using markdown.'
DEFAULT_GPT_MODEL = "gpt-3.5-turbo"
EMBEDDING_MODEL = "text-embedding-ada-002"
EMBED_DIM = 1536


def token_limits(model: str = DEFAULT_GPT_MODEL) -> int:
    """Return the maximum number of tokens for a model."""
    return {
        "gpt-3.5-turbo": 4096,
        "gpt-4": 8192,
        "gpt-4-32k": 32768,
        "text-embedding-ada-002": 8191,
    }[model]


def count_message_tokens(messages: List[Message],
                         model: str = "gpt-3.5-turbo-0301") -> int:
    """
    Returns the number of tokens used by a list of messages.

    Args:
        messages (list): A list of messages, each of which is a dictionary
            containing the role and content of the message.
        model (str): The name of the model to use for tokenization.
            Defaults to "gpt-3.5-turbo-0301".

    Returns:
        int: The number of tokens used by the list of messages.
    """
    try:
        encoding = tiktoken.encoding_for_model(model)
    except KeyError:
        logger.warn("Warning: model not found. Using cl100k_base encoding.")
        encoding = tiktoken.get_encoding("cl100k_base")
    if model == "gpt-3.5-turbo":
        # !Note: gpt-3.5-turbo may change over time.
        # Returning num tokens assuming gpt-3.5-turbo-0301.")
        return count_message_tokens(messages, model="gpt-3.5-turbo-0301")
    elif model == "gpt-4":
        # !Note: gpt-4 may change over time. Returning num tokens assuming gpt-4-0314.")
        return count_message_tokens(messages, model="gpt-4-0314")
    elif model == "gpt-3.5-turbo-0301":
        tokens_per_message = (
            4  # every message follows <|start|>{role/name}\n{content}<|end|>\n
        )
        tokens_per_name = -1  # if there's a name, the role is omitted
    elif model == "gpt-4-0314":
        tokens_per_message = 3
        tokens_per_name = 1
    else:
        raise NotImplementedError(
            f"num_tokens_from_messages() is not implemented for model {model}.\n"
            " See https://github.com/openai/openai-python/blob/main/chatml.md for"
            " information on how messages are converted to tokens.")
    num_tokens = 0
    for message in messages:
        num_tokens += tokens_per_message
        for key, value in message.items():
            num_tokens += len(encoding.encode(value))
            if key == "name":
                num_tokens += tokens_per_name
    num_tokens += 3  # every reply is primed with <|start|>assistant<|message|>
    return num_tokens


def count_string_tokens(string: str, model_name: str) -> int:
    """
    Returns the number of tokens in a text string.

    Args:
        string (str): The text string.
        model_name (str): The name of the encoding to use. (e.g., "gpt-3.5-turbo")

    Returns:
        int: The number of tokens in the text string.
    """
    encoding = tiktoken.encoding_for_model(model_name)
    return len(encoding.encode(string))


def create_chat_message(role, content) -> Message:
    """
    Create a chat message with the given role and content.

    Args:
    role (str): The role of the message sender, e.g., "system", "user", or "assistant".
    content (str): The content of the message.

    Returns:
    dict: A dictionary containing the role and content of the message.
    """
    return {"role": role, "content": content}


def generate_context(prompt,
                     relevant_memory,
                     full_message_history,
                     model,
                     summary=None):
    current_context = [
        create_chat_message("system", prompt),
        create_chat_message(
            "system", f"The current time and date is {time.strftime('%c')}"),
        create_chat_message(
            "system",
            f"This reminds you of these events from your past:\n{relevant_memory}\n\n",
        ),
    ]
    if summary is not None:
        current_context.append(
            create_chat_message(
                "system",
                f"This is a summary of the conversation so far:\n{summary}\n\n"
            ))

    # Add messages from the full message history until we reach the token limit
    next_message_to_add_index = len(full_message_history) - 1
    insertion_index = len(current_context)
    # Count the currently used tokens
    current_tokens_used = count_message_tokens(current_context, model)
    return (
        next_message_to_add_index,
        current_tokens_used,
        insertion_index,
        current_context,
    )


@tenacity.retry(wait=tenacity.wait_exponential(multiplier=1, min=4, max=10),
                stop=tenacity.stop_after_attempt(5),
                retry=tenacity.retry_if_exception_type(
                    (RateLimitError, APIError, Timeout,
                     ServiceUnavailableError, APIConnectionError)))
def create_chat_completion(
    messages: List[Message],  # type: ignore
    model: Optional[str] = None,
    temperature: float = 0.9,
    max_tokens: Optional[int] = None,
) -> str:
    """Create a chat completion using the OpenAI API

    Args:
        messages (List[Message]): The messages to send to the chat completion
        model (str, optional): The model to use. Defaults to None.
        temperature (float, optional): The temperature to use. Defaults to 0.9.
        max_tokens (int, optional): The max tokens to use. Defaults to None.

    Returns:
        str: The response from the chat completion
    """
    response = openai.ChatCompletion.create(
        model=model,
        messages=messages,
        temperature=temperature,
        max_tokens=max_tokens,
    )
    try:
        resp = response.choices[0].message["content"]
    except:
        try:
            return response.error.message
        except:
            logger.error(f"Error in create_chat_completion: {response}")
        raise
    return resp


@tenacity.retry(wait=tenacity.wait_exponential(multiplier=1, min=4, max=10),
                stop=tenacity.stop_after_attempt(5),
                retry=tenacity.retry_if_exception_type(
                    (RateLimitError, APIError, Timeout,
                     ServiceUnavailableError, APIConnectionError)))
def get_embedding(
    text: str,
    *_,
    model: str = EMBEDDING_MODEL,
    **kwargs,
) -> List[float]:
    """Create an embedding using the OpenAI API

    Args:
        text (str): The text to embed.
        kwargs: Other arguments to pass to the OpenAI API embedding creation call.

    Returns:
        List[float]: The embedding.
    """
    return openai.Embedding.create(
        model=model,
        input=[text],
        **kwargs,
    )["data"][0]["embedding"]


def chat_with_ai(prompt,
                 user_input,
                 full_message_history,
                 permanent_memory,
                 summary=None,
                 model=DEFAULT_GPT_MODEL,
                 token_limit=None):
    """Interact with the OpenAI API, sending the prompt, user input, message history,
    and permanent memory.

    Args:
        prompt (str): The prompt explaining the rules to the AI.
        user_input (str): The input from the user.
        full_message_history (list): The list of all messages sent between the
            user and the AI.
        permanent_memory (Obj): The memory object containing the permanent
            memory.
        summary (str): The summary of the conversation so far.
        model (str): The name of the model to use for tokenization.
        token_limit (int): The maximum number of tokens allowed in the API call.

    Returns:
        str: The AI's response.
    """

    # Reserve 1000 tokens for the response

    if token_limit is None:
        token_limit = token_limits(model)

    logger.debug(f"Token limit: {token_limit}")
    send_token_limit = token_limit - 1000
    if len(full_message_history) == 0:
        relevant_memory = ""
    else:
        recent_history = full_message_history[-5:]
        shuffle(recent_history)
        relevant_memories = permanent_memory.get_relevant(
            str(recent_history), 5)
        if relevant_memories:
            shuffle(relevant_memories)
        relevant_memory = str(relevant_memories)

    logger.debug(f"Memory Stats: {permanent_memory.get_stats()}")

    (
        next_message_to_add_index,
        current_tokens_used,
        insertion_index,
        current_context,
    ) = generate_context(prompt, relevant_memory, full_message_history, model,
                         summary)

    while current_tokens_used > 2500:
        # remove memories until we are under 2500 tokens
        relevant_memory = relevant_memory[:-1]
        (
            next_message_to_add_index,
            current_tokens_used,
            insertion_index,
            current_context,
        ) = generate_context(prompt, relevant_memory, full_message_history,
                             model, summary)

    current_tokens_used += count_message_tokens(
        [create_chat_message("user", user_input)],
        model)  # Account for user input (appended later)

    while next_message_to_add_index >= 0:
        # print (f"CURRENT TOKENS USED: {current_tokens_used}")
        message_to_add = full_message_history[next_message_to_add_index]

        tokens_to_add = count_message_tokens([message_to_add], model)
        if current_tokens_used + tokens_to_add > send_token_limit:
            break

        # Add the most recent message to the start of the current context,
        #  after the two system prompts.
        current_context.insert(insertion_index,
                               full_message_history[next_message_to_add_index])

        # Count the currently used tokens
        current_tokens_used += tokens_to_add

        # Move to the next most recent message in the full message history
        next_message_to_add_index -= 1

    # Append user input, the length of this is accounted for above
    current_context.extend([create_chat_message("user", user_input)])

    # Calculate remaining tokens
    tokens_remaining = token_limit - current_tokens_used
    # assert tokens_remaining >= 0, "Tokens remaining is negative.

    # TODO: use a model defined elsewhere, so that model can contain
    # temperature and other settings we care about
    assistant_reply = create_chat_completion(
        model=model,
        messages=current_context,
        max_tokens=tokens_remaining,
    )

    # Update full message history
    full_message_history.append(create_chat_message("user", user_input))
    full_message_history.append(
        create_chat_message("assistant", assistant_reply))

    return assistant_reply


def create_default_embeddings():
    return np.zeros((0, EMBED_DIM)).astype(np.float32)


@dataclasses.dataclass
class CacheContent:
    texts: List[str] = dataclasses.field(default_factory=list)
    embeddings: np.ndarray = dataclasses.field(
        default_factory=create_default_embeddings)


class LocalCache():
    """A class that stores the memory in a local file"""

    def __init__(self, path=None) -> None:
        """Initialize a class instance

        Args:
            path: str

        Returns:
            None
        """
        if path is None:
            self.filename = None
        else:
            self.filename = Path(path)
            self.filename.touch(exist_ok=True)
        try:
            with open(self.filename, 'rb') as f:
                self.data = pickle.load(f)
        except:
            self.data = CacheContent()

    def add(self, text: str):
        """
        Add text to our list of texts, add embedding as row to our
            embeddings-matrix

        Args:
            text: str

        Returns: None
        """
        if "Command Error:" in text:
            return ""
        self.data.texts.append(text)

        embedding = get_embedding(text)

        vector = np.array(embedding).astype(np.float32)
        vector = vector[np.newaxis, :]
        self.data.embeddings = np.concatenate(
            [
                self.data.embeddings,
                vector,
            ],
            axis=0,
        )

        if self.filename is not None:
            with open(self.filename, "wb") as f:
                pickle.dump(self.data, f)
        return text

    def clear(self) -> str:
        """
        Clears the data in memory.

        Returns: A message indicating that the memory has been cleared.
        """
        self.data = CacheContent()
        return "Obliviated"

    def get(self, data: str) -> list[Any] | None:
        """
        Gets the data from the memory that is most relevant to the given data.

        Args:
            data: The data to compare to.

        Returns: The most relevant data.
        """
        return self.get_relevant(data, 1)

    def get_relevant(self, text: str, k: int) -> list[Any]:
        """ "
        matrix-vector mult to find score-for-each-row-of-matrix
         get indices for top-k winning scores
         return texts for those indices
        Args:
            text: str
            k: int

        Returns: List[str]
        """
        if self.data.embeddings.shape[0] == 0:
            return []
        embedding = get_embedding(text)

        scores = np.dot(self.data.embeddings, embedding)

        top_k_indices = np.argsort(scores)[-k:][::-1]

        return [self.data.texts[i] for i in top_k_indices]

    def get_stats(self) -> tuple[int, tuple[int, ...]]:
        """
        Returns: The stats of the local cache.
        """
        return len(self.data.texts), self.data.embeddings.shape


class Completion():

    def __init__(self,
                 system_prompt=DEFAULT_SYSTEM_PROMPT,
                 model=DEFAULT_GPT_MODEL):
        self.messages = [{"role": "system", "content": system_prompt}]
        self.title = 'untitled'
        self.last_time = datetime.now()
        self.completion = None
        self.total_tokens = 0
        self.prompt_tokens = 0
        self.completion_tokens = 0
        self.model = model

    def make_title(self):

        text = [
            f'{d["role"]} :\n"""\n{d["content"]}\n"""'
            for d in self.messages[1:]
        ]

        messages = [{
            "role": "system",
            "content": 'You are a helpful assistant.'
        }, {
            'role':
            "user",
            'content': ("总结以下对话的内容并为其取个标题以概括对话的内容，标题长度不超过100个字符。"
                        "不得包含`?:*,<>\\/` 等不能用于文件路径的字符。"
                        "返回的结果除了标题本身，不要包含额外的内容，省略结尾的句号。\n" + '\n\n'.join(text))
        }]
        completion = openai.ChatCompletion.create(model=self.model,
                                                  messages=messages)
        content = completion.choices[0].message['content']
        return f"{time.strftime('%Y%m%d%H%M')} {content}"

    def say(self, msg):
        self.last_time = datetime.now()
        self.messages.append({"role": "user", "content": msg})
        self.completion = openai.ChatCompletion.create(model=self.model,
                                                       messages=self.messages)
        self.total_tokens += self.completion.usage.total_tokens
        self.completion_tokens += self.completion.usage.completion_tokens
        self.prompt_tokens += self.completion.usage.prompt_tokens
        message = self.completion.choices[0].message
        self.messages.append({
            "role": message['role'],
            "content": message['content']
        })
        return message['content']

    def save(self):
        if self.title == 'untitled':
            self.title = self.make_title()

        filepath = Path.home() / 'chatGPT' / f"{self.title}.completion"
        filepath.parent.mkdir(parents=True, exist_ok=True)
        with open(filepath, 'wb') as f:
            pickle.dump(self, f)


class Conversation():

    def __init__(self,
                 system_prompt=DEFAULT_SYSTEM_PROMPT,
                 model=DEFAULT_GPT_MODEL):
        self.system_prompt = system_prompt
        self.summary = None
        self.history = []
        self.memory = LocalCache()
        self.title = None
        self.last_time = datetime.now()
        self.model = model
        self._pool = ThreadPoolExecutor()
        self._save_future = None

    def __del__(self):
        if self._save_future is not None:
            self._save_future.result()
        self._pool.shutdown()

    def _validate_title(self, title: str) -> str:
        title.replace('\\/:.*?%&#\"\'<>{}|\n\r\t_', ' ')
        title = title.strip()
        title = '_'.join(title.split())
        if len(title) > 70:
            title = title[:70]
        while title[-1] in ' .。,，-_':
            title = title[:-1]
        return title

    def make_title(self):
        messages = [{
            "role": "system",
            "content": 'You are a helpful assistant.'
        }]

        tokens = count_string_tokens(messages[0]['content'], self.model)

        query = ("请根据以下对话内容，总结出中文标题，长度不超过100个字符。"
                 "请注意，标题必须是合法的文件名，省略结尾的句号。返回结果不得包含额外的解释和格式。\n"
                 "对话内容：\n")

        text = []
        for msg in self.history:
            text.append(f'{msg["role"]} :\n<quote>{msg["content"]}</quote>')
            tokens += count_string_tokens(query + '\n'.join(text), self.model)
            if tokens > token_limits(self.model) - 500:
                text.pop()
                break
        messages.append({"role": "user", "content": query + '\n'.join(text)})

        try:
            self.last_time = datetime.now()
            content = create_chat_completion(messages, self.model)
            title = self._validate_title(content)
            return f"{time.strftime('%Y%m%d%H%M%S')} {title}"
        except:
            return f"{time.strftime('%Y%m%d%H%M%S')} untitled"

    def ask(self, query):
        self.last_time = datetime.now()

        reply = chat_with_ai(self.system_prompt, query, self.history,
                             self.memory, self.summary, self.model,
                             token_limits(self.model))
        return reply

    def _save(self):
        if len(self.history) == 0:
            return

        if self.title is None:
            self.title = self.make_title()

        filepath = Path.home() / 'chatGPT' / f"{self.title}.conversation"
        filepath.parent.mkdir(parents=True, exist_ok=True)
        with open(filepath, 'wb') as f:
            pickle.dump(self, f)

    def save(self):
        self._save_future = self._pool.submit(self._save)
        return self._save_future

    def __getstate__(self):
        state = self.__dict__.copy()
        del state['_pool']
        del state['_save_future']
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self._pool = ThreadPoolExecutor(max_workers=1)
        self._save_future = None


ipy = get_ipython()

current_completion = Conversation()


def chat(line, cell):
    global current_completion
    if line:
        args = line.split()
        current_completion.save()
        model = DEFAULT_GPT_MODEL
        if args[0] in ['gpt-4', 'gpt-3.5', 'gpt-3.5-turbo']:
            model = args[0]
            if model == 'gpt-3.5':
                model = 'gpt-3.5-turbo'
            if len(args) > 1:
                prompt = ' '.join(args[1:])
            else:
                prompt = DEFAULT_SYSTEM_PROMPT
        current_completion = Conversation(system_prompt=prompt, model=model)
        if args[0] in ['end', 'save', 'bye']:
            return
    content = current_completion.ask(cell)
    display(Markdown(content))
    ipy.set_next_input('%%chat\n')


def autosave_completion():
    global current_completion
    if (datetime.now() - current_completion.last_time).seconds > 300 and len(
            current_completion.history) >= 3:
        current_completion.save()
    elif len(current_completion.history) > 7:
        current_completion.save()


def load_chat(index):
    global current_completion

    filepath = Path.home() / 'chatGPT'
    if not filepath.exists():
        return
    for i, f in enumerate(
            sorted(filepath.glob('*.conversation'),
                   key=lambda f: f.stat().st_mtime,
                   reverse=True)):
        if i == index:
            if current_completion is not None:
                current_completion.save().result()
            with open(f, 'rb') as f:
                current_completion = pickle.load(f)
            break


def show_chat(index=None):
    if index is not None:
        load_chat(index)
    messages = current_completion.history
    for msg in messages:
        display(Markdown(f"**{msg['role']}**\n\n{msg['content']}"))
    ipy.set_next_input('%%chat\n')


def list_chat():
    filepath = Path.home() / 'chatGPT'
    if not filepath.exists():
        return
    rows = ["|index|title|length|time|", "|:---:|:---:|:---:|:---:|"]
    for i, f in enumerate(
            sorted(filepath.glob('*.conversation'),
                   key=lambda f: f.stat().st_mtime,
                   reverse=True)):
        with open(f, 'rb') as f:
            completion = pickle.load(f)
        rows.append(
            f"|{i}|{completion.title}|{len(completion.history)}|{completion.last_time}|"
        )
    display(Markdown('\n'.join(rows)))


if ipy is not None:
    ipy.register_magic_function(chat, 'cell', magic_name='chat')
    ipy.events.register('post_run_cell', autosave_completion)



import hashlib
import zlib
from pathlib import Path

DATAPATH = Path.home() / 'data'
CHUNKSIZE = 1024 * 1024 * 4  # 4 MB


def set_data_path(base_path: str) -> None:
    global DATAPATH
    DATAPATH = Path(base_path)


def get_data_path() -> Path:
    return DATAPATH


def save_chunk(data: bytes, compressed: bool = False) -> tuple[str, str]:
    if compressed:
        data = zlib.compress(data)
    hashstr = hashlib.sha1(data).hexdigest()
    file = get_data_path(
    ) / 'chunks' / hashstr[:2] / hashstr[2:4] / hashstr[4:]
    file.parent.mkdir(parents=True, exist_ok=True)
    with open(file, 'wb') as f:
        f.write(data)
    return str('/'.join(file.parts[-4:])), len(data)


def load_chunk(file: str, compressed: bool = False) -> bytes:
    if file.startswith('chunks/'):
        with open(get_data_path() / file, 'rb') as f:
            data = f.read()
    elif file.startswith('packs/'):
        *filepath, start, size = file.split('/')
        filepath = '/'.join(filepath)
        with open(get_data_path() / filepath, 'rb') as f:
            f.seek(int(start))
            data = f.read(int(size))
    else:
        raise ValueError('Invalid file path: ' + file)
    if compressed:
        data = zlib.decompress(data)
    return data


def pack_chunk(pack: str, chunkfile: str) -> str:
    pack = get_data_path() / 'packs' / pack
    pack.parent.mkdir(parents=True, exist_ok=True)
    with open(pack, 'ab') as f:
        buf = load_chunk(chunkfile)
        start = f.tell()
        size = len(buf)
        f.write(buf)
    return str('/'.join(pack.parts[-2:])) + '/' + str(start) + '/' + str(size)


def delete_chunk(file: str):
    file = get_data_path() / file
    file.unlink()



#
# Packet format
# -------------
#
# Offset    Length   Notes
# ------    ------   -----
#
#   0       1        Operation code with 1 being request and 2 being response
#   1       1        Hardware type with 1 being "Ethernet 10Mb"
#   2       1        Hardware address length with 6 for Ethernet
#   3       1        Hops - usually 0 unless DHCP relaying in operation
#   4-7     4        Transaction ID (selected randomly by client)
#   8-9     2        Seconds - might be used by a server to prioritise requests
#  10-11    2        Flags (only most significan bit used for broadcast)
#  12-15    4        Client Internet address (might be requested by client)
#  16-19    4        Your Internet address (the IP assigned by the server)
#  20-23    4        Server Internet address (the IP of the server)
#  24-27    4        Gateway Internet address (if DHCP relaying in operation)
#  28-43    16       Client hardware address - only first 6 bytes used for Ethernet
#  44-107   64       Text name of server (optional)
# 108-235   128      Boot file name (optional - used for PXE booting)
# 236-239   4        Magic cookie (decimal values 99, 130, 83, 99 )
#

#
# DHCP option codes
# -----------------
#
# 1   - Subnet mask
# 3   - Router(s)
# 6   - DNS name server(s)
# 12  - Hostname
# 15  - Domain name
# 28  - Broadcast address
# 33  - Static route
# 42  - Network time protocol servers
# 51  - IP address lease time
# 53  - DHCP Message Type (DHCPDISCOVER, DHCPOFFER, etc)
# 54  - Server identifier
# 55  - Parameter Request List
# 57  - Maximum DHCP Message Size
# 58  - Renewal (T1) time value
# 59  - Renewal (T2) time value
# 60  - Vendor Class Identifier
# 61  - Client Identifier
# 67  - Boot file name (e.g. PXE booting)
# 80  -
# 116 -
# 119 -
# 145 -
# 167 -
# 171 -
#

import asyncio
import logging
import socket
from typing import NamedTuple

from waveforms.sys.net.dhcp import (MAGIC_COOKIE, DHCPMessageType, DHCPOption,
                                    DHCPPacket)

logger = logging.getLogger(__name__)


class DHCPServerProtocol:

    def connection_made(self, transport):
        self.transport = transport
        address, port = self.transport.get_extra_info('sockname')
        self.server_ip = '192.168.254.1'

    def connection_lost(self, exc):
        pass

    def datagram_received(self, data, addr):
        try:
            received = DHCPPacket.decode(data)
        except ValueError as e:
            logger.error("Error decoding DHCP packet: %s", e)
            return

        print("\nReceived DHCP packet from %s:%d" % (addr[0], addr[1]))
        print(received)

        # see if there is a DHCP message type option
        if 53 not in received.options:
            return

        messagetype = received.get_option(DHCPOption.DHCP_MSG_TYPE)
        if (messagetype != DHCPMessageType.DISCOVER) and (
                messagetype != DHCPMessageType.REQUEST):
            logger.error(
                "ignoring: DHCP message type not supported by this implementation"
            )
            return

        try:
            assigned_ip, subnet, gateway = get_ip_from_mac(received.chaddr)
        except:
            return

        response = DHCPPacket()
        response.op = 2
        response.xid = received.xid
        response.flags = received.flags
        response.chaddr = received.chaddr
        response.yiaddr = assigned_ip
        response.siaddr = self.server_ip

        if messagetype == DHCPMessageType.DISCOVER:
            # DHCPOFFER
            response.set_option(DHCPOption.DHCP_MSG_TYPE,
                                DHCPMessageType.OFFER)
        elif messagetype == DHCPMessageType.REQUEST:
            # DHCPACK
            response.set_option(DHCPOption.DHCP_MSG_TYPE, DHCPMessageType.ACK)
        else:
            return

        lease_time = 86400
        response.set_option(DHCPOption.SUBNET_MASK, subnet)
        response.set_option(DHCPOption.ROUTER, gateway)
        response.set_option(DHCPOption.IP_LEASE_TIME, lease_time)
        response.set_option(DHCPOption.SERVER_ID, self.server_ip)

        response_packet = response.encode()

        # send it
        print("\nSending DHCP packet to %s:%d" % (addr[0], addr[1]))
        print(DHCPPacket.decode(response_packet))
        self.send_packet(response_packet, addr)

    def send_packet(self, packet, addr):
        if addr == ('0.0.0.0', 68):
            send_address = ('255.255.255.255', 68)
        else:
            send_address = addr
        self.transport.sendto(packet, send_address)


class IPConfig(NamedTuple):
    ip: str
    subnet: str
    gateway: str


ip_table = {
    '00:19:AF:04:92:C4':
    IPConfig(ip='192.168.254.150', subnet='255.255.255.0', gateway='0.0.0.0')
}


def get_ip_from_mac(macaddr):
    ip_config = ip_table.get(macaddr, None)
    return ip_config.ip, ip_config.subnet, ip_config.gateway


async def main(server_ip='0.0.0.0'):
    loop = asyncio.get_running_loop()

    transport, protocol = await loop.create_datagram_endpoint(
        lambda: DHCPServerProtocol(),
        local_addr=(server_ip, 67),
        allow_broadcast=True)

    # print('start', server_ip)

    try:
        await asyncio.sleep(3600)  # Serve for 1 hour.
    finally:
        transport.close()


if __name__ == '__main__':
    asyncio.run(main(server_ip='0.0.0.0'))



import math

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.colors import LogNorm, Normalize, SymLogNorm
from matplotlib.ticker import EngFormatter, LogFormatterSciNotation
from scipy.interpolate import griddata


def good_for_logscale(x, threshold=4):
    if np.any(x <= 0):
        return False
    mid = (np.nanmin(x) + np.nanmax(x)) / 2
    a = np.count_nonzero(np.nan_to_num(x <= mid, nan=0))
    b = np.count_nonzero(np.nan_to_num(x >= mid, nan=0))
    if a / b > threshold:
        return True
    return False


def equal_logspace(x):
    logx = np.logspace(np.log10(x[0]), np.log10(x[-1]), len(x))
    return np.allclose(x, logx)


def equal_linspace(x):
    linearx = np.linspace(x[0], x[-1], len(x))
    return np.allclose(x, linearx)


def as_1d_data(x, y, z):
    x = np.asarray(x)
    y = np.asarray(y)
    z = np.asarray(z)
    if z.ndim == 1:
        return x, y, z

    if z.ndim == 2:
        x, y = np.meshgrid(x, y)
        return x.ravel(), y.ravel(), z.ravel()

    raise ValueError("z must be 1D or 2D")


def griddata_logx_logy(x, y, z, shape=(401, 401)):
    x, y, z = as_1d_data(x, y, z)
    xspace = np.logspace(np.log10(x.min()), np.log10(x.max()), shape[0])
    yspace = np.logspace(np.log10(y.min()), np.log10(y.max()), shape[1])
    xgrid, ygrid = np.meshgrid(xspace, yspace)
    zgrid = griddata((x, y), z, (xgrid, ygrid), method='nearest')
    return xspace, yspace, zgrid


def griddata_logx_linear_y(x, y, z, shape=(401, 401)):
    x, y, z = as_1d_data(x, y, z)
    xspace = np.logspace(np.log10(x.min()), np.log10(x.max()), shape[0])
    yspace = np.linspace(y.min(), y.max(), shape[1])
    xgrid, ygrid = np.meshgrid(xspace, yspace)
    zgrid = griddata((x, y), z, (xgrid, ygrid), method='nearest')
    return xspace, yspace, zgrid


def griddata_linear_x_logy(x, y, z, shape=(401, 401)):
    x, y, z = as_1d_data(x, y, z)
    xspace = np.linspace(x.min(), x.max(), shape[0])
    yspace = np.logspace(np.log10(y.min()), np.log10(y.max()), shape[1])
    xgrid, ygrid = np.meshgrid(xspace, yspace)
    zgrid = griddata((x, y), z, (xgrid, ygrid), method='nearest')
    return xspace, yspace, zgrid


def griddata_linear_x_linear_y(x, y, z, shape=(401, 401)):
    x, y, z = as_1d_data(x, y, z)
    xspace = np.linspace(x.min(), x.max(), shape[0])
    yspace = np.linspace(y.min(), y.max(), shape[1])
    xgrid, ygrid = np.meshgrid(xspace, yspace)
    zgrid = griddata((x, y), z, (xgrid, ygrid), method='nearest')
    return xspace, yspace, zgrid


def _get_log_ticks(x):
    log10x = np.log10(x)

    major_ticks = np.array(
        range(math.floor(log10x[0]) - 1,
              math.ceil(log10x[-1]) + 1))
    minor_ticks = np.hstack([
        np.log10(np.linspace(2, 10, 9, endpoint=False)) + x
        for x in major_ticks
    ])

    major_ticks = major_ticks[(major_ticks >= log10x[0]) *
                              (major_ticks <= log10x[-1])]
    minor_ticks = minor_ticks[(minor_ticks >= log10x[0]) *
                              (minor_ticks <= log10x[-1])]

    return log10x, major_ticks, minor_ticks


class MyLogFormatter(EngFormatter):

    def format_ticks(self, values):
        if self.unit is None or self.unit == '':
            fmt = LogFormatterSciNotation()
            return [f"${fmt.format_data(10.0**x)}$" for x in values]
        else:
            return super().format_ticks(values)

    def format_eng(self, x):
        if self.unit is None or self.unit == '':
            self.unit = ''
            return f"{10.0**x:g}"
        else:
            return super().format_eng(10.0**x)


def imshow_logx(x, y, z, x_unit=None, ax=None, **kwargs):
    if ax is None:
        ax = plt.gca()

    log10x, major_ticks, minor_ticks = _get_log_ticks(x)

    dlogx, dy = log10x[1] - log10x[0], y[1] - y[0]
    extent = (log10x[0] - dlogx / 2, log10x[-1] + dlogx / 2, y[0] - dy / 2,
              y[-1] + dy / 2)

    img = ax.imshow(z, extent=extent, **kwargs)

    ax.set_xticks(major_ticks, minor=False)
    ax.xaxis.set_major_formatter(MyLogFormatter(x_unit))
    ax.set_xticks(minor_ticks, minor=True)

    return img


def imshow_logy(x, y, z, y_unit=None, ax=None, **kwargs):
    if ax is None:
        ax = plt.gca()

    log10y, major_ticks, minor_ticks = _get_log_ticks(y)

    dlogy, dx = log10y[1] - log10y[0], x[1] - x[0]
    extent = (x[0] - dx / 2, x[-1] + dx / 2, log10y[0] - dlogy / 2,
              log10y[-1] + dlogy / 2)

    img = ax.imshow(z, extent=extent, **kwargs)

    ax.set_yticks(major_ticks, minor=False)
    ax.yaxis.set_major_formatter(MyLogFormatter(y_unit))
    ax.set_yticks(minor_ticks, minor=True)

    return img


def imshow_loglog(x, y, z, x_unit=None, y_unit=None, ax=None, **kwargs):
    if ax is None:
        ax = plt.gca()

    log10x, x_major_ticks, x_minor_ticks = _get_log_ticks(x)
    log10y, y_major_ticks, y_minor_ticks = _get_log_ticks(y)

    dlogx, dlogy = log10x[1] - log10x[0], log10y[1] - log10y[0]
    extent = (log10x[0] - dlogx / 2, log10x[-1] + dlogx / 2,
              log10y[0] - dlogy / 2, log10y[-1] + dlogy / 2)

    img = ax.imshow(z, extent=extent, **kwargs)

    ax.set_xticks(x_major_ticks, minor=False)
    ax.xaxis.set_major_formatter(MyLogFormatter(x_unit))
    ax.set_xticks(x_minor_ticks, minor=True)

    ax.set_yticks(y_major_ticks, minor=False)
    ax.yaxis.set_major_formatter(MyLogFormatter(y_unit))
    ax.set_yticks(y_minor_ticks, minor=True)

    return img


def plot_lines(x,
               y,
               z,
               xlabel,
               ylabel,
               zlabel,
               x_unit,
               y_unit,
               z_unit,
               ax,
               xscale='linear',
               yscale='linear',
               zscale='linear',
               index=None,
               **kwds):
    z = np.asarray(z)
    if len(y) > len(x):
        x, y = y, x
        xlabel, ylabel = ylabel, xlabel
        xscale, yscale = yscale, xscale
        z = z.T
    if index is not None:
        y = y[index]
        z = z[index, :]

    for i, l in enumerate(y):
        if y_unit:
            label = f"{ylabel}={l:.3} [{y_unit}]"
        else:
            if isinstance(l, float):
                label = f"{ylabel}={l:.3}"
            else:
                label = f"{ylabel}={l}"
        ax.plot(x, z[i, :], label=label, **kwds)
    ax.legend()
    xlabel = f"{xlabel} [{x_unit}]" if x_unit else xlabel
    zlabel = f"{zlabel} [{z_unit}]" if z_unit else zlabel
    ax.set_xlabel(xlabel)
    ax.set_ylabel(zlabel)
    ax.set_xscale(xscale)
    ax.set_yscale(zscale)


def plot_img(x,
             y,
             z,
             xlabel,
             ylabel,
             zlabel,
             x_unit,
             y_unit,
             z_unit,
             fig,
             ax,
             xscale='linear',
             yscale='linear',
             zscale='linear',
             resolution=None,
             **kwds):
    kwds.setdefault('origin', 'lower')
    kwds.setdefault('aspect', 'auto')
    kwds.setdefault('interpolation', 'nearest')

    vmin = kwds.get('vmin', np.nanmin(z))
    vmax = kwds.get('vmax', np.nanmax(z))
    if zscale == 'log':
        kwds.setdefault('norm', LogNorm(vmax=vmax, vmin=vmin))
    elif zscale == 'symlog':
        kwds.setdefault('norm', SymLogNorm(vmax=vmax,
                                           vmin=vmin,
                                           linthresh=1e-5))
    else:
        kwds.setdefault('norm', Normalize(vmin=vmin, vmax=vmax))
    zlabel = f"{zlabel} [{z_unit}]" if z_unit else zlabel

    band_area = False
    if x.ndim == 1 and y.ndim == 2 and y.shape[1] == x.shape[0]:
        x = np.asarray([x] * y.shape[0])
        band_area = True
    elif x.ndim == 2 and y.ndim == 1 and x.shape[0] == y.shape[0]:
        y = np.asarray([y] * x.shape[1]).T
        band_area = True
    if band_area:
        kwds.pop('origin', None)
        kwds.pop('aspect', None)
        kwds.pop('interpolation', None)
        pc = ax.pcolormesh(x, y, z, **kwds)
        xlabel = f"{xlabel} [{x_unit}]" if x_unit else xlabel
        ylabel = f"{ylabel} [{y_unit}]" if y_unit else ylabel
        ax.set_xlabel(xlabel)
        ax.set_ylabel(ylabel)
        cb = fig.colorbar(pc, ax=ax)
        ax.set_xscale(xscale)
        ax.set_yscale(yscale)
        cb.set_label(zlabel)
        return

    if resolution is None:
        resolution = (401, 401)
    elif isinstance(resolution, int):
        resolution = (resolution, resolution)

    if (z.ndim == 1 or (xscale == 'linear' and not equal_linspace(x))
            or (yscale == 'linear' and not equal_linspace(y))
            or (xscale == 'log' and not equal_logspace(x))
            or (yscale == 'log' and not equal_logspace(y))):
        griddata = {
            ('log', 'log'): griddata_logx_logy,
            ('log', 'linear'): griddata_logx_linear_y,
            ('linear', 'log'): griddata_linear_x_logy,
            ('linear', 'linear'): griddata_linear_x_linear_y,
        }[(xscale, yscale)]
        x, y, z = griddata(x, y, z, resolution)

    if (xscale, yscale) == ('linear', 'linear'):
        dx, dy = x[1] - x[0], y[1] - y[0]
        extent = (x[0] - dx / 2, x[-1] + dx / 2, y[0] - dy / 2, y[-1] + dy / 2)
        kwds.setdefault('extent', extent)
        img = ax.imshow(np.asarray(z), **kwds)
        xlabel = f"{xlabel} [{x_unit}]" if x_unit else xlabel
        ylabel = f"{ylabel} [{y_unit}]" if y_unit else ylabel
    elif (xscale, yscale) == ('log', 'linear'):
        ylabel = f"{ylabel} [{y_unit}]" if y_unit else ylabel
        img = imshow_logx(x, y, z, x_unit, ax, **kwds)
    elif (xscale, yscale) == ('linear', 'log'):
        xlabel = f"{xlabel} [{x_unit}]" if x_unit else xlabel
        img = imshow_logy(x, y, z, y_unit, ax, **kwds)
    elif (xscale, yscale) == ('log', 'log'):
        img = imshow_loglog(x, y, z, x_unit, y_unit, ax, **kwds)
    else:
        pass
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    cb = fig.colorbar(img, ax=ax, norm=kwds.get('norm', None))
    cb.set_label(zlabel)


def plot_scatter(x,
                 y,
                 z,
                 xlabel,
                 ylabel,
                 zlabel,
                 x_unit,
                 y_unit,
                 z_unit,
                 fig,
                 ax,
                 xscale='linear',
                 yscale='linear',
                 zscale='linear',
                 **kwds):
    if np.any(np.iscomplex(z)):
        s = np.abs(z)
        c = np.angle(z)
    else:
        s = np.abs(z)
        c = z.real
    ax.scatter(x, y, s=s, c=c, **kwds)
    xlabel = f"{xlabel} [{x_unit}]" if x_unit else xlabel
    ylabel = f"{ylabel} [{y_unit}]" if y_unit else ylabel
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    ax.set_xscale(xscale)
    ax.set_yscale(yscale)


def autoplot(x,
             y,
             z,
             xlabel='x',
             ylabel='y',
             zlabel='z',
             x_unit='',
             y_unit='',
             z_unit='',
             fig=None,
             ax=None,
             index=None,
             xscale='auto',
             yscale='auto',
             zscale='auto',
             max_lines=3,
             scatter_lim=1000,
             resolution=None,
             **kwds):
    """
    Plot a 2D array as a line plot or an image.

    Parameters:
        x (array): x values
        y (array): y values
        z (array): z values
        xlabel (str): x label
        ylabel (str): y label
        zlabel (str): z label
        x_unit (str): x unit
        y_unit (str): y unit
        z_unit (str): z unit
        fig (Figure): figure to plot on
        ax (Axes): axes to plot on
        index (int): index of the line to plot
        xscale (str): x scale 'auto', 'linear' or 'log'
        yscale (str): y scale 'auto', 'linear' or 'log'
        zscale (str): z scale 'auto', 'linear' or 'log'
        max_lines (int): maximum number of lines to plot
        **kwds: keyword arguments passed to plot_img or plot_lines
    """
    if ax is not None:
        fig = ax.figure
    if fig is None:
        fig = plt.gcf()
    if ax is None:
        ax = fig.add_subplot(111)

    x = np.asarray(x)
    y = np.asarray(y)
    z = np.asarray(z)

    if xscale == 'auto':
        if good_for_logscale(x):
            xscale = 'log'
        else:
            xscale = 'linear'
    if yscale == 'auto':
        if good_for_logscale(y):
            yscale = 'log'
        else:
            yscale = 'linear'
    if zscale == 'auto':
        if good_for_logscale(z):
            zscale = 'log'
        else:
            zscale = 'linear'

    if x.shape == y.shape == z.shape and z.size < scatter_lim:
        plot_scatter(x,
                     y,
                     z,
                     xlabel,
                     ylabel,
                     zlabel,
                     x_unit,
                     y_unit,
                     z_unit,
                     fig,
                     ax,
                     xscale=xscale,
                     yscale=yscale,
                     zscale=zscale,
                     **kwds)
    elif z.ndim == 2 and (len(y) <= max_lines or len(x) <= max_lines
                          or index is not None):
        plot_lines(x,
                   y,
                   z,
                   xlabel,
                   ylabel,
                   zlabel,
                   x_unit=x_unit,
                   y_unit=y_unit,
                   z_unit=z_unit,
                   xscale=xscale,
                   yscale=yscale,
                   zscale=zscale,
                   ax=ax,
                   index=index,
                   **kwds)
    else:
        plot_img(x,
                 y,
                 z,
                 xlabel,
                 ylabel,
                 zlabel,
                 x_unit=x_unit,
                 y_unit=y_unit,
                 z_unit=z_unit,
                 xscale=xscale,
                 yscale=yscale,
                 zscale=zscale,
                 fig=fig,
                 ax=ax,
                 resolution=resolution,
                 **kwds)



import itertools

import matplotlib.pyplot as plt
import numpy as np


def cycle(x0, y0, r, a=0, b=2 * np.pi):
    t = np.linspace(a, b, 101)
    x = r * np.cos(t) + x0
    y = r * np.sin(t) + y0
    return x, y


def plotSquare(xc, yc, xs, ys, radius=0, color='C0', ax=None):
    ax = plt.gca() if ax is None else ax

    l, r, b, u = xc - xs / 2, xc + xs / 2, yc - ys / 2, yc + ys / 2
    x = np.linspace(l + radius, r - radius, 2)
    y = np.linspace(b + radius, u - radius, 2)

    ax.plot(x, [u, u], color=color)
    ax.plot(x, [b, b], color=color)
    ax.plot([l, l], y, color=color)
    ax.plot([r, r], y, color=color)

    xx, yy = cycle(x[-1], y[-1], radius, 0, np.pi / 2)
    ax.plot(xx, yy, color=color)

    xx, yy = cycle(x[0], y[-1], radius, np.pi / 2, np.pi)
    ax.plot(xx, yy, color=color)

    xx, yy = cycle(x[0], y[0], radius, np.pi, np.pi * 3 / 2)
    ax.plot(xx, yy, color=color)

    xx, yy = cycle(x[-1], y[0], radius, np.pi * 3 / 2, np.pi * 2)
    ax.plot(xx, yy, color=color)


def plotMesure(x0=0, y0=0, size=1, color='C0', ax=None):

    ax = plt.gca() if ax is None else ax

    x, y = cycle(x0, y0 - 0.37 * size, 0.57 * size, np.pi / 4, np.pi * 3 / 4)
    ax.plot(x, y, color=color)

    plotSquare(x0, y0, size, 0.8 * size, radius=0.2 * size, color=color, ax=ax)

    angle = np.pi / 6

    ax.arrow(x0,
             y0 - 0.25 * size,
             0.55 * size * np.sin(angle),
             0.55 * size * np.cos(angle),
             head_width=0.07 * size,
             head_length=0.1 * size,
             color=color)


def plot_seq(ax, waves, measure=[1], gap=4, maxTime=20, xlim=None):
    t = np.linspace(0, maxTime, 1001)
    tt = np.linspace(0, maxTime - gap, 1001)
    styles = ['-', '--', ':', '-.']
    colors = ['C0', 'C1', 'C2', 'C3', 'C4', 'C5', 'C6', 'C7', 'C8', 'C9']
    for i, (wavgroup, color) in enumerate(zip(waves, itertools.cycle(colors))):
        if i in measure:
            time = tt
        else:
            time = t
        if not isinstance(wavgroup, tuple):
            wavgroup = (wavgroup, )

        for wav, style in zip(wavgroup, itertools.cycle(styles)):
            ax.plot(time, wav(time) - gap * i, style, color=color)
            if i in measure:
                plotMesure(maxTime - gap / 2,
                           -gap * i,
                           gap,
                           color='black',
                           ax=ax)

    ax.axis('equal')

    ax.spines['left'].set_visible(False)
    ax.spines['right'].set_visible(False)
    ax.spines['bottom'].set_visible(False)
    ax.spines['top'].set_visible(False)
    ax.set_xticks([])
    ax.set_yticks([])
    if xlim is not None:
        ax.set_xlim(xlim)



import asyncio
import copy
import inspect
import itertools
import lzma
import os
import pickle
import platform
import re
import subprocess
import sys
import uuid
from concurrent.futures import ProcessPoolExecutor
from graphlib import TopologicalSorter
from pathlib import Path
from typing import Any, Awaitable, Callable, Iterable

import dill
import numpy as np
import zmq

from ..sys.rpc.zmq_socket import ZMQContextManager
from .expression import Env, Expression, Symbol
from .optimize import NgOptimizer
from .record import Record
from .server import default_record_port
from .space import Optimizer, OptimizeSpace, Space
from .utils import async_zip, call_function, dump_globals

try:
    from tqdm.notebook import tqdm
except:

    class tqdm():

        def update(self, n):
            pass

        def close(self):
            pass

        def reset(self):
            pass


__process_uuid = uuid.uuid1()
__task_counter = itertools.count()
__notebook_id = None

if os.getenv('QULAB_SERVER'):
    default_server = os.getenv('QULAB_SERVER')
else:
    default_server = f'tcp://127.0.0.1:{default_record_port}'
if os.getenv('QULAB_EXECUTOR'):
    default_executor = os.getenv('QULAB_EXECUTOR')
else:
    default_executor = default_server


def yapf_reformat(cell_text):
    try:
        import isort
        import yapf.yapflib.yapf_api

        fname = f"f{uuid.uuid1().hex}"

        def wrap(source):
            lines = [f"async def {fname}():"]
            for line in source.split('\n'):
                lines.append("    " + line)
            return '\n'.join(lines)

        def unwrap(source):
            lines = []
            for line in source.split('\n'):
                if line.startswith(f"async def {fname}():"):
                    continue
                lines.append(line[4:])
            return '\n'.join(lines)

        cell_text = re.sub('^%', '#%#', cell_text, flags=re.M)
        reformated_text = unwrap(
            yapf.yapflib.yapf_api.FormatCode(wrap(isort.code(cell_text)))[0])
        return re.sub('^#%#', '%', reformated_text, flags=re.M)
    except:
        return cell_text


def get_installed_packages():
    result = subprocess.run([sys.executable, '-m', 'pip', 'freeze'],
                            stdout=subprocess.PIPE,
                            text=True)

    lines = result.stdout.split('\n')
    packages = []
    for line in lines:
        if line:
            packages.append(line)
    return packages


def get_system_info():
    info = {
        'OS': platform.uname()._asdict(),
        'Python': sys.version,
        'PythonExecutable': sys.executable,
        'PythonPath': sys.path,
        'packages': get_installed_packages()
    }
    return info


def current_notebook():
    return __notebook_id


async def create_notebook(name: str, database=default_server, socket=None):
    global __notebook_id

    async with ZMQContextManager(zmq.DEALER, connect=database,
                                 socket=socket) as socket:
        await socket.send_pyobj({'method': 'notebook_create', 'name': name})
        __notebook_id = await socket.recv_pyobj()


async def save_input_cells(notebook_id,
                           input_cells,
                           database=default_server,
                           socket=None):
    async with ZMQContextManager(zmq.DEALER, connect=database,
                                 socket=socket) as socket:
        await socket.send_pyobj({
            'method': 'notebook_extend',
            'notebook_id': notebook_id,
            'input_cells': input_cells
        })
        return await socket.recv_pyobj()


async def create_config(config: dict, database=default_server, socket=None):
    async with ZMQContextManager(zmq.DEALER, connect=database,
                                 socket=socket) as socket:
        buf = lzma.compress(pickle.dumps(config))
        await socket.send_pyobj({'method': 'config_update', 'update': buf})
        return await socket.recv_pyobj()


async def get_config(config_id: int, database=default_server, socket=None):
    async with ZMQContextManager(zmq.DEALER, connect=database,
                                 socket=socket) as socket:
        await socket.send_pyobj({
            'method': 'config_get',
            'config_id': config_id
        })
        buf = await socket.recv_pyobj()
        return pickle.loads(lzma.decompress(buf))


def task_uuid():
    return uuid.uuid3(__process_uuid, str(next(__task_counter)))


def _get_depends(func: Callable):
    try:
        sig = inspect.signature(func)
    except:
        return []

    args = []
    for name, param in sig.parameters.items():
        if param.kind in [
                param.POSITIONAL_ONLY, param.POSITIONAL_OR_KEYWORD,
                param.KEYWORD_ONLY
        ]:
            args.append(name)
        elif param.kind == param.VAR_KEYWORD:
            pass
        elif param.kind == param.VAR_POSITIONAL:
            raise ValueError('not support VAR_POSITIONAL')
    return args


class Promise():
    __slots__ = ['task', 'key', 'attr']

    def __init__(self, task, key=None, attr=None):
        self.task = task
        self.key = key
        self.attr = attr

    def __await__(self):

        async def _getitem(task, key):
            return (await task)[key]

        async def _getattr(task, attr):
            return getattr(await task, attr)

        if self.key is not None:
            return _getitem(self.task, self.key).__await__()
        elif self.attr is not None:
            return _getattr(self.task, self.attr).__await__()
        else:
            return self.task.__await__()

    def __getitem__(self, key):
        return Promise(self.task, key, None)

    def __getattr__(self, attr):
        return Promise(self.task, None, attr)


def _run_function_in_process(buf):
    func, args, kwds = dill.loads(buf)
    return func(*args, **kwds)


class Scan():

    def __new__(cls, *args, mixin=None, **kwds):
        if mixin is None:
            return super().__new__(cls)
        for k in dir(mixin):
            if not hasattr(cls, k):
                try:
                    setattr(cls, k, getattr(mixin, k))
                except:
                    pass
        return super().__new__(cls)

    def __init__(self,
                 app: str = 'task',
                 tags: tuple[str] = (),
                 database: str | Path
                 | None = default_server,
                 dump_globals: bool = False,
                 max_workers: int = 4,
                 max_promise: int = 100,
                 mixin=None):
        self.id = task_uuid()
        self.record = None
        self.config = None
        self.description = {
            'app': app,
            'tags': tags,
            'config': None,
            'loops': {},
            'intrinsic_loops': {},
            'consts': {},
            'functions': {},
            'getters': {},
            'setters': {},
            'optimizers': {},
            'namespace': {} if dump_globals else None,
            'actions': {},
            'dependents': {},
            'order': {},
            'axis': {},
            'independent_variables': set(),
            'filters': {},
            'total': {},
            'database': database,
            'hiden': ['self', r'^__.*', r'.*__$'],
            'entry': {
                'system': get_system_info(),
                'env': {},
                'shell': '',
                'cmds': [],
                'scripts': []
            },
        }
        self._current_level = 0
        self._variables = {}
        self._main_task = None
        self._sock = None
        self._sem = asyncio.Semaphore(max_promise + 1)
        self._bar: dict[int, tqdm] = {}
        self._hide_pattern_re = re.compile('|'.join(self.description['hiden']))
        self._msg_queue = asyncio.Queue()
        self._prm_queue = asyncio.Queue()
        self._single_step = True
        self._max_workers = max_workers
        self._max_promise = max_promise
        self._executors = ProcessPoolExecutor(max_workers=max_workers)

    def __del__(self):
        try:
            self._main_task.cancel()
        except:
            pass

    def __getstate__(self) -> dict:
        state = self.__dict__.copy()
        del state['record']
        del state['_sock']
        del state['_main_task']
        del state['_bar']
        del state['_msg_queue']
        del state['_prm_queue']
        del state['_sem']
        del state['_executors']
        return state

    def __setstate__(self, state: dict) -> None:
        self.__dict__.update(state)
        self.record = None
        self._sock = None
        self._main_task = None
        self._bar = {}
        self._prm_queue = asyncio.Queue()
        self._msg_queue = asyncio.Queue()
        self._sem = asyncio.Semaphore(self._max_promise + 1)
        self._executors = ProcessPoolExecutor(max_workers=self._max_workers)
        for opt in self.description['optimizers'].values():
            opt.scanner = self

    def __del__(self):
        try:
            self._main_task.cancel()
        except:
            pass
        self._executors.shutdown()

    @property
    def current_level(self):
        return self._current_level

    @property
    def variables(self) -> dict[str, Any]:
        return self._variables

    async def _emit(self, current_level, step, position, variables: dict[str,
                                                                         Any]):
        for key, value in list(variables.items()):
            if inspect.isawaitable(value) and not self.hiden(key):
                variables[key] = await value
        if self._sock is not None:
            await self._sock.send_pyobj({
                'task': self.id,
                'method': 'record_append',
                'record_id': self.record.id,
                'level': current_level,
                'step': step,
                'position': position,
                'variables': {
                    k: v
                    for k, v in variables.items() if not self.hiden(k)
                }
            })
        else:
            self.record.append(current_level, step, position, {
                k: v
                for k, v in variables.items() if not self.hiden(k)
            })

    def emit(self, current_level, step, position, variables: dict[str, Any]):
        self._msg_queue.put_nowait(
            asyncio.create_task(
                self._emit(current_level, step, position, variables.copy())))

    def hide(self, name: str):
        self.description['hiden'].append(name)
        self._hide_pattern_re = re.compile('|'.join(self.description['hiden']))

    def hiden(self, name: str) -> bool:
        return bool(self._hide_pattern_re.match(name))

    async def _filter(self, variables: dict[str, Any], level: int = 0):
        try:
            return all(await asyncio.gather(*[
                call_function(fun, variables) for fun in itertools.chain(
                    self.description['filters'].get(level, []),
                    self.description['filters'].get(-1, []))
            ]))
        except:
            return True

    async def create_record(self):
        if self._sock is not None:
            await self._sock.send_pyobj({
                'task':
                self.id,
                'method':
                'record_create',
                'description':
                dill.dumps(self.description)
            })

            record_id = await self._sock.recv_pyobj()
            return Record(record_id, self.description['database'],
                          self.description)
        return Record(None, self.description['database'], self.description)

    def get(self, name: str):
        if name in self.description['consts']:
            return self.description['consts'][name]
        else:
            return Symbol(name)

    def _add_search_space(self, name: str, level: int, space):
        if level not in self.description['loops']:
            self.description['loops'][level] = []
        self.description['loops'][level].append((name, space))

    def add_depends(self, name: str, depends: list[str]):
        if isinstance(depends, str):
            depends = [depends]
        if name not in self.description['dependents']:
            self.description['dependents'][name] = set()
        self.description['dependents'][name].update(depends)

    def add_filter(self, func: Callable, level: int = -1):
        """
        Add a filter function to the scan.

        Args:
            func: A callable object or an instance of Expression.
            level: The level of the scan to add the filter. -1 means any level.
        """
        if level not in self.description['filters']:
            self.description['filters'][level] = []
        self.description['filters'][level].append(func)

    def set(self,
            name: str,
            value,
            depends: Iterable[str] | None = None,
            setter: Callable | None = None):
        try:
            dill.dumps(value)
        except:
            raise ValueError('value is not serializable.')
        if isinstance(value, Expression):
            self.add_depends(name, value.symbols())
            self.description['functions'][name] = value
        elif callable(value):
            if depends:
                self.add_depends(name, depends)
                s = ','.join(depends)
                self.description['functions'][f'_tmp_{name}'] = value
                self.description['functions'][name] = eval(
                    f"lambda self, {s}: self.description['functions']['_tmp_{name}']({s})"
                )
            else:
                self.add_depends(name, _get_depends(value))
                self.description['functions'][name] = value
        else:
            try:
                value = Space.fromarray(value)
            except:
                pass
            self.description['consts'][name] = value
        if setter:
            self.description['setters'][name] = setter

    def search(self,
               name: str,
               space: Iterable | Expression | Callable | OptimizeSpace,
               level: int | None = None,
               setter: Callable | None = None,
               intrinsic: bool = False):
        if level is not None:
            if not intrinsic:
                assert level >= 0, 'level must be greater than or equal to 0.'
        if intrinsic:
            assert isinstance(space, (np.ndarray, list, tuple, range, Space)), \
                'space must be an instance of np.ndarray, list, tuple, range or Space.'
            self.description['intrinsic_loops'][name] = level
            self.set(name, space)
        elif isinstance(space, OptimizeSpace):
            space.name = name
            space.optimizer.dimensions[name] = space.space
            self._add_search_space(name, space.optimizer.level, space)
            self.add_depends(space.optimizer.name, [name])
        else:
            if level is None:
                raise ValueError('level must be provided.')
            try:
                space = Space.fromarray(space)
            except:
                pass
            self._add_search_space(name, level, space)
            if isinstance(space, Expression) or callable(space):
                self.add_depends(name, space.symbols())
        if setter:
            self.description['setters'][name] = setter

    def trace(self,
              name: str,
              depends: list[str],
              getter: Callable | None = None):
        self.add_depends(name, depends)
        if getter:
            self.description['getters'][name] = getter

    def minimize(self,
                 name: str,
                 level: int,
                 method=NgOptimizer,
                 maxiter=100,
                 getter: Callable | None = None,
                 **kwds) -> Optimizer:
        assert level >= 0, 'level must be greater than or equal to 0.'
        opt = Optimizer(self,
                        name,
                        level,
                        method,
                        maxiter,
                        minimize=True,
                        **kwds)
        self.description['optimizers'][name] = opt
        if getter:
            self.description['getters'][name] = getter
        return opt

    def maximize(self,
                 name: str,
                 level: int,
                 method=NgOptimizer,
                 maxiter=100,
                 getter: Callable | None = None,
                 **kwds) -> Optimizer:
        assert level >= 0, 'level must be greater than or equal to 0.'
        opt = Optimizer(self,
                        name,
                        level,
                        method,
                        maxiter,
                        minimize=False,
                        **kwds)
        self.description['optimizers'][name] = opt
        if getter:
            self.description['getters'][name] = getter
        return opt

    async def _update_progress(self):
        while True:
            task = await self._prm_queue.get()
            await task
            self._prm_queue.task_done()

    async def _send_msg(self):
        while True:
            task = await self._msg_queue.get()
            await task
            self._msg_queue.task_done()

    async def run(self):
        assymbly(self.description)
        if self.config:
            self.description['config'] = await create_config(
                self.config, self.description['database'], self._sock)
        if current_notebook() is None:
            await create_notebook('untitle', self.description['database'],
                                  self._sock)
        cell_id = await save_input_cells(current_notebook(),
                                         self.description['entry']['scripts'],
                                         self.description['database'],
                                         self._sock)
        self.description['entry']['scripts'] = cell_id
        if isinstance(
                self.description['database'],
                str) and self.description['database'].startswith("tcp://"):
            async with ZMQContextManager(
                    zmq.DEALER,
                    connect=self.description['database']) as socket:
                self._sock = socket
                await self._run()
        else:
            await self._run()

    async def _run(self):
        send_msg = asyncio.create_task(self._send_msg())
        update_progress_task = asyncio.create_task(self._update_progress())

        self._variables = {'self': self}

        consts = {}
        for k, v in self.description['consts'].items():
            if isinstance(v, Space):
                consts[k] = v.toarray()
            else:
                consts[k] = v

        await update_variables(self._variables, consts,
                               self.description['setters'])
        for level, total in self.description['total'].items():
            if total == np.inf:
                total = None
            self._bar[level] = tqdm(total=total)

        updates = await call_many_functions(
            self.description['order'].get(-1, []),
            self.description['functions'], self.variables)
        await update_variables(self.variables, updates,
                               self.description['setters'])

        self.record = await self.create_record()
        await self.work()
        for level, bar in self._bar.items():
            bar.close()

        if self._single_step:
            self.variables.update(await call_many_functions(
                self.description['order'].get(-1, []),
                self.description['getters'], self.variables))

            self.emit(0, 0, 0, self.variables)
            self.emit(-1, 0, 0, {})

        await self._prm_queue.join()
        update_progress_task.cancel()
        await self._msg_queue.join()
        send_msg.cancel()
        return self.variables

    async def done(self):
        if self._main_task is not None:
            try:
                await self._main_task
            except asyncio.CancelledError:
                pass

    def finished(self):
        return self._main_task.done()

    def start(self):
        import asyncio
        self._main_task = asyncio.create_task(self.run())

    async def submit(self, server=default_executor):
        assymbly(self.description)
        async with ZMQContextManager(zmq.DEALER, connect=server) as socket:
            await socket.send_pyobj({
                'method': 'submit',
                'description': dill.dumps(self.description)
            })
            self.id = await socket.recv_pyobj()
            await socket.send_pyobj({'method': 'get_record_id', 'id': self.id})
            record_id = await socket.recv_pyobj()
            self.record = Record(record_id, self.description['database'],
                                 self.description)

    def cancel(self):
        if self._main_task is not None:
            self._main_task.cancel()

    async def _reset_progress_bar(self, level):
        if level in self._bar:
            self._bar[level].reset()

    async def _update_progress_bar(self, level, n: int):
        if level in self._bar:
            self._bar[level].update(n)

    async def iter(self, **kwds):
        if self.current_level >= len(self.description['loops']):
            return
        step = 0
        position = 0
        self._prm_queue.put_nowait(self._reset_progress_bar(
            self.current_level))
        async for variables in _iter_level(
                self.variables,
                self.description['loops'].get(self.current_level, []),
                self.description['order'].get(self.current_level, []),
                self.description['functions'], self.description['optimizers'],
                self.description['setters'], self.description['getters']):
            self._current_level += 1
            if await self._filter(variables, self.current_level - 1):
                yield variables
                self._single_step = False
                self.emit(self.current_level - 1, step, position, variables)
                step += 1
            position += 1
            self._current_level -= 1
            self._prm_queue.put_nowait(
                self._update_progress_bar(self.current_level, 1))
        if self.current_level == 0:
            self.emit(self.current_level - 1, 0, 0, {})
            for name, value in self.variables.items():
                if inspect.isawaitable(value):
                    self.variables[name] = await value
            await self._prm_queue.join()

    async def work(self, **kwds):
        if self.current_level in self.description['actions']:
            action = self.description['actions'][self.current_level]
            coro = action(self, **kwds)
            if inspect.isawaitable(coro):
                await coro
        else:
            async for variables in self.iter(**kwds):
                await self.do_something(**kwds)

    async def do_something(self, **kwds):
        await self.work(**kwds)

    def mount(self, action: Callable, level: int):
        """
        Mount a action to the scan.

        Args:
            action: A callable object.
            level: The level of the scan to mount the action.
        """
        self.description['actions'][level] = action

    async def promise(self, awaitable: Awaitable | Callable, *args,
                      **kwds) -> Promise:
        """
        Promise to calculate asynchronous function and return the result in future.

        Args:
            awaitable: An awaitable object.

        Returns:
            Promise: A promise object.
        """
        if inspect.isawaitable(awaitable):
            async with self._sem:
                task = asyncio.create_task(self._await(awaitable))
                self._prm_queue.put_nowait(task)
                return Promise(task)
        elif inspect.iscoroutinefunction(awaitable):
            return await self.promise(awaitable(*args, **kwds))
        elif callable(awaitable):
            try:
                buf = dill.dumps((awaitable, args, kwds))
                task = asyncio.get_running_loop().run_in_executor(
                    self._executors, _run_function_in_process, buf)
                self._prm_queue.put_nowait(task)
                return Promise(task)
            except:
                return awaitable(*args, **kwds)
        else:
            return awaitable

    async def _await(self, awaitable: Awaitable):
        async with self._sem:
            return await awaitable


def assymbly(description):
    import __main__
    from IPython import get_ipython

    if isinstance(description['namespace'], dict):
        description['namespace'] = dump_globals()

    ipy = get_ipython()
    if ipy is not None:
        description['entry']['shell'] = 'ipython'
        description['entry']['scripts'] = [
            yapf_reformat(cell_text) for cell_text in ipy.user_ns['In']
        ]
    else:
        try:
            description['entry']['shell'] = 'shell'
            description['entry']['cmds'] = [
                sys.executable, __main__.__file__, *sys.argv[1:]
            ]
            description['entry']['scripts'] = []
            try:
                with open(__main__.__file__) as f:
                    description['entry']['scripts'].append(f.read())
            except:
                pass
        except:
            pass

    description['entry']['env'] = {k: v for k, v in os.environ.items()}

    mapping = {
        label: level
        for level, label in enumerate(
            sorted(
                set(description['loops'].keys())
                | {k
                   for k in description['actions'].keys() if k >= 0}))
    }

    if -1 in description['actions']:
        mapping[-1] = max(mapping.values()) + 1

    levels = sorted(mapping.values())
    for k in description['actions'].keys():
        if k < -1:
            mapping[k] = levels[k]

    description['loops'] = dict(
        sorted([(mapping[k], v) for k, v in description['loops'].items()]))
    description['actions'] = {
        mapping[k]: v
        for k, v in description['actions'].items()
    }

    for level, loops in description['loops'].items():
        description['total'][level] = np.inf
        for name, space in loops:
            try:
                description['total'][level] = min(description['total'][level],
                                                  len(space))
            except:
                pass

    dependents = copy.deepcopy(description['dependents'])

    for level in levels:
        range_list = description['loops'].get(level, [])
        if level > 0:
            if f'#__loop_{level}' not in description['dependents']:
                dependents[f'#__loop_{level}'] = set()
            dependents[f'#__loop_{level}'].add(f'#__loop_{level-1}')
        for name, _ in range_list:
            if name not in description['dependents']:
                dependents[name] = set()
            dependents[name].add(f'#__loop_{level}')

    def _get_all_depends(key, graph):
        ret = set()
        if key not in graph:
            return ret

        for e in graph[key]:
            ret.update(_get_all_depends(e, graph))
        ret.update(graph[key])
        return ret

    full_depends = {}
    for key in dependents:
        full_depends[key] = _get_all_depends(key, dependents)

    levels = {}
    passed = set()
    all_keys = set(description['consts'].keys())
    for key in dependents:
        all_keys.add(key)
        all_keys.update(dependents[key])
    for level in reversed(description['loops'].keys()):
        tag = f'#__loop_{level}'
        for key, deps in full_depends.items():
            all_keys.update(deps)
            all_keys.add(key)
            if key.startswith('#__loop_'):
                continue
            if tag in deps:
                if level not in levels:
                    levels[level] = set()
                if key not in passed:
                    passed.add(key)
                    levels[level].add(key)
    levels[-1] = {
        key
        for key in all_keys - passed if not key.startswith('#__loop_')
    }

    order = []
    ts = TopologicalSorter(dependents)
    ts.prepare()
    while ts.is_active():
        ready = ts.get_ready()
        order.append(ready)
        for k in ready:
            ts.done(k)

    description['order'] = {}

    for level in sorted(levels):
        keys = set(levels[level])
        description['order'][level] = []
        for ready in order:
            ready = list(keys & set(ready))
            if ready:
                description['order'][level].append(ready)
                keys -= set(ready)

    axis = {}
    independent_variables = set(description['intrinsic_loops'].keys())

    for name in description['consts']:
        axis[name] = ()
    for level, range_list in description['loops'].items():
        for name, iterable in range_list:
            if isinstance(iterable, OptimizeSpace):
                axis[name] = tuple(range(level + 1))
                continue
            elif isinstance(iterable, (np.ndarray, list, tuple, range, Space)):
                independent_variables.add(name)
            axis[name] = (level, )

    for level, group in description['order'].items():
        for names in group:
            for name in names:
                if name not in description['dependents']:
                    if name not in axis:
                        axis[name] = (level, )
                else:
                    d = set()
                    for n in description['dependents'][name]:
                        d.update(axis[n])
                    if name not in axis:
                        axis[name] = tuple(sorted(d))
                    else:
                        axis[name] = tuple(sorted(set(axis[name]) | d))
    description['axis'] = axis
    description['independent_variables'] = independent_variables

    return description


async def update_variables(variables: dict[str, Any], updates: dict[str, Any],
                           setters: dict[str, Callable]):
    coros = []
    for name, value in updates.items():
        if name in setters:
            coro = setters[name](value)
            if inspect.isawaitable(coro):
                coros.append(coro)
        variables[name] = value
    if coros:
        await asyncio.gather(*coros)


async def _iter_level(variables,
                      iters: list[tuple[str, Iterable | Expression | Callable
                                        | OptimizeSpace]],
                      order: list[list[str]],
                      functions: dict[str, Callable | Expression],
                      optimizers: dict[str, Optimizer],
                      setters: dict[str, Callable] = {},
                      getters: dict[str, Callable] = {}):
    iters_d = {}
    env = Env()
    env.variables = variables
    opts = {}

    for name, iter in iters:
        if isinstance(iter, OptimizeSpace):
            if iter.optimizer.name not in opts:
                opts[iter.optimizer.name] = iter.optimizer.create()
        elif isinstance(iter, Expression):
            iters_d[name] = iter.eval(env)
        elif isinstance(iter, Space):
            iters_d[name] = iter.toarray()
        elif callable(iter):
            iters_d[name] = await call_function(iter, variables)
        else:
            iters_d[name] = iter

    maxiter = 0xffffffff
    for name, opt in opts.items():
        opt_cfg = optimizers[name]
        maxiter = min(maxiter, opt_cfg.maxiter)

    async for args in async_zip(*iters_d.values(), range(maxiter)):
        await update_variables(variables, dict(zip(iters_d.keys(), args[:-1])),
                               setters)
        for name, opt in opts.items():
            args = opt.ask()
            opt_cfg = optimizers[name]
            await update_variables(variables, {
                n: v
                for n, v in zip(opt_cfg.dimensions.keys(), args)
            }, setters)

        await update_variables(
            variables, await call_many_functions(order, functions, variables),
            setters)

        yield variables

        variables.update(await call_many_functions(order, getters, variables))

        for name, opt in opts.items():
            opt_cfg = optimizers[name]
            args = [variables[n] for n in opt_cfg.dimensions.keys()]
            if name not in variables:
                raise ValueError(f'{name} not in variables.')
            fun = variables[name]
            if inspect.isawaitable(fun):
                fun = await fun
            if opt_cfg.minimize:
                opt.tell(args, fun)
            else:
                opt.tell(args, -fun)

    for name, opt in opts.items():
        opt_cfg = optimizers[name]
        result = opt.get_result()
        await update_variables(
            variables, {
                name: value
                for name, value in zip(opt_cfg.dimensions.keys(), result.x)
            }, setters)
        variables[name] = result.fun
    if opts:
        yield variables


async def call_many_functions(order: list[list[str]],
                              functions: dict[str, Callable],
                              variables: dict[str, Any]) -> dict[str, Any]:
    ret = {}
    for group in order:
        waited = []
        coros = []
        for name in group:
            if name in functions:
                waited.append(name)
                coros.append(call_function(functions[name], variables | ret))
        if coros:
            results = await asyncio.gather(*coros)
            ret.update(dict(zip(waited, results)))
    return ret



from datetime import date, datetime
from typing import Sequence

import dill
from sqlalchemy import Column, DateTime, ForeignKey, Integer, String
from sqlalchemy.orm import Session, relationship

from .base import Base
from .config import Config
from .tag import get_object_with_tags, has_tags, tag


@has_tags
class Record(Base):
    __tablename__ = 'records'

    id = Column(Integer, primary_key=True)
    ctime = Column(DateTime, default=datetime.utcnow)
    mtime = Column(DateTime, default=datetime.utcnow)
    atime = Column(DateTime, default=datetime.utcnow)
    name = Column(String)
    datafile_id = Column(Integer, ForeignKey('files.id'))
    configfile_id = Column(Integer, ForeignKey('files.id'))
    meta_id = Column(Integer, ForeignKey('files.id'))

    datafile = relationship("File", foreign_keys=[datafile_id])
    configfile = relationship("File", foreign_keys=[configfile_id])
    metafile = relationship("File", foreign_keys=[meta_id])

    @property
    def data(self) -> dict:
        result = dill.loads(self.datafile.read())
        result['meta'] = self.meta
        result['meta']['config'] = self.config
        return result

    @data.setter
    def data(self, data: dict):
        meta = data.pop('meta', {})
        self.meta = meta
        self.datafile.write(dill.dumps(data))

    @property
    def config(self) -> dict:
        return dill.loads(self.configfile.read())

    @config.setter
    def config(self, data: dict):
        self.configfile.write(dill.dumps(data))

    @property
    def meta(self) -> dict:
        meta = dill.loads(self.metafile.read())
        meta['id'] = self.id
        meta['name'] = self.name
        meta['ctime'] = self.ctime
        meta['mtime'] = self.mtime
        meta['atime'] = self.atime
        return meta

    @meta.setter
    def meta(self, data: dict):
        if 'config' in data:
            config = data.pop('config', {})
            self.config = config
        self.metafile.write(dill.dumps(data))

    def export(self, path):
        with open(path, 'wb') as f:
            f.write(self.data)

    def __init__(self, name: str, data: bytes) -> None:
        self.name = name
        self.data = data

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return f'Record({self.name!r})'


def query_record(
        session: Session,
        offset: int = 0,
        limit: int = 10,
        name: str | None = None,
        tags: Sequence[str] = (),
        before: datetime | date | None = None,
        after: datetime | date | None = None) -> tuple[int, dict, dict]:
    from waveforms.dicttree import foldDict

    local_tm = datetime.fromtimestamp(0)
    utc_tm = datetime.utcfromtimestamp(0)
    tz_offset = local_tm - utc_tm
    table = {'header': ['ID', 'Name', 'tags', 'created time'], 'body': []}
    name_lst = sorted(
        set([
            n for n, *_ in get_object_with_tags(session.query(Record.name),
                                                Record, *tags).all()
        ]))
    name_tr = foldDict(dict([(name, None) for name in name_lst]))

    query = get_object_with_tags(session, Record, *tags)

    if name is not None:
        if name.endswith('*'):
            query = query.filter(Record.name.like(name[:-1] + '%'))
        else:
            query = query.filter(Record.name == name)
    if before is not None:
        if isinstance(before, date):
            before = datetime(before.year, before.month, before.day)
        query = query.filter(Record.ctime <= before - tz_offset)
    if after is not None:
        if isinstance(after, date):
            after = datetime(after.year, after.month, after.day)
        query = query.filter(Record.ctime >= after - tz_offset)
    total = query.count()
    for r in query.order_by(Record.ctime.desc()).limit(limit).offset(offset):
        tags = sorted([t.text for t in r.tags])
        ctime = r.ctime + tz_offset
        row = [r.id, r.name, tags, ctime]
        table['body'].append(row)

    return total, name_tr, table


def update_tags(session: Session,
                record_id: int,
                tags: Sequence[str],
                append: bool = False):
    record = session.get(Record, record_id)
    if record is None:
        return False
    if append:
        old = [t.text for t in record.tags]
        for t in old:
            if t not in tags:
                tags.append(t)
    record.tags = [tag(session, t) for t in tags]
    try:
        session.commit()
    except Exception:
        session.rollback()
        return False
    return True


def remove_tags(session: Session, record_id: int, tags: Sequence[str]):
    record = session.get(Record, record_id)
    if record is None:
        return False
    old = [t.text for t in record.tags]
    record.tags = [tag(session, t) for t in old if t not in tags]
    try:
        session.commit()
    except Exception:
        session.rollback()
        return False
    return True



import matplotlib.pyplot as plt
import numpy as np


class DataPicker():

    def __init__(self, ax=None):
        self.points_and_text = {}
        self.points = None
        self.hline = None
        self.vline = None
        self.text = None
        if ax is None:
            ax = plt.gca()
        self.ax = ax
        self.ax.figure.canvas.mpl_connect('button_press_event', self.on_click)
        # self.ax.figure.canvas.mpl_connect('motion_notify_event', self.on_move)
        self.ax.figure.canvas.mpl_connect('key_press_event', self.on_key_press)
        self.mode = 'pick'

    def on_key_press(self, event):
        if event.key == 'a':
            if self.mode != 'pick':
                self.mode = 'pick'
            else:
                self.mode = 'default'

    def on_move(self, event):
        if event.inaxes is self.ax:
            # self.hline = self.ax.axhline(y=np.nan, color='r', lw=1)
            # self.vline = self.ax.axvline(x=np.nan, color='r', lw=1)
            # self.text = self.ax.text(0, 0, '', verticalalignment='center')
            self.hline.set_ydata(event.ydata)
            self.vline.set_xdata(event.xdata)
            self.text.set_position((event.xdata, event.ydata))
            self.text.set_text(f'({event.xdata:.2f}, {event.ydata:.2f})')
            self.ax.draw()

    def on_click(self, event):
        if self.mode != 'pick':
            return
        # 鼠标左键的button值为1
        if event.button == 1 and event.inaxes is self.ax:
            point = (event.xdata, event.ydata)
            text = self.ax.text(point[0],
                                point[1],
                                f'({point[0]:.2f}, {point[1]:.2f})',
                                verticalalignment='center')
            self.points_and_text[point] = text
            x, y = self.get_xy()
            if self.points is None:
                self.points, = self.ax.plot(x, y, 'ro')
            else:
                self.points.set_data(x, y)
            self.ax.draw()

        elif event.button == 3 and event.inaxes is self.ax:
            for point, text in list(self.points_and_text.items()):
                point_xdisplay, point_ydisplay = self.ax.transData.transform_point(
                    point)

                distance = np.sqrt((point_xdisplay - event.x)**2 +
                                   (point_ydisplay - event.y)**2)
                if distance < 10:
                    text.remove()
                    self.points_and_text.pop(point)
                    if self.points_and_text:
                        x, y = self.get_xy()
                        self.points.set_data(x, y)
                    else:
                        self.points.remove()
                        self.points = None
                    self.ax.draw()
                    break

    def get_xy(self):
        if self.points_and_text:
            data = np.asarray(list(self.points_and_text.keys()))
            x, y = data[:, 0], data[:, 1]

            index = np.argsort(x)
            x = np.asarray(x)[index]
            y = np.asarray(y)[index]
            return x, y
        else:
            return np.array([]), np.array([])



from typing import Optional

import zmq
import zmq.asyncio
import zmq.auth
from watchdog.events import FileSystemEventHandler
from watchdog.observers import Observer
from zmq.auth.asyncio import AsyncioAuthenticator
from zmq.auth.thread import ThreadAuthenticator


class ReloadCertificatesHandler(FileSystemEventHandler):

    def __init__(self, zmq: 'ZMQContextManager'):
        self.zmq = zmq

    def on_modified(self, event):
        self.zmq.reload_certificates()


class ZMQContextManager:
    """
    A context manager for managing ZeroMQ sockets with asynchronous support.
    It handles the creation, connection, binding, and security configuration of
    the sockets and ensures proper resource cleanup.

    The context manager can be used as a synchronous context manager or an
    asynchronous context manager. When used as an asynchronous context manager,
    the socket is created with an asyncio context and can be used with the
    `await` keyword.

    The security settings for the socket can be configured using the secret key
    and public key parameters. If the secret key is provided, the socket is
    configured to use ZeroMQ curve encryption. The public key of the server can
    also be provided for client sockets to connect to a server with curve
    encryption. You can also provide the paths to the secret key and public key
    files, and the public key of the server to load the keys from files. The
    keys can be reloaded automatically when the files are modified by setting
    the `public_keys_location` parameter.

    To generate a secret key and public key pair, you can use the following
    commands:
    
        ```bash
        # Generate a secret key and public key pair
        python -c "import zmq.auth; zmq.auth.create_certificates('.', 'filename')"
        ```

    Attributes:
        socket_type: zmq.SocketType
            The type of the socket to create (e.g., zmq.REP).
        bind: str, optional
            The address to bind the socket to.
        connect: str, optional
            The address to connect the socket to.
        secret_key_file: str, optional
            The path to the secret key file for ZeroMQ curve encryption.
        server_public_key_file: str, optional
            The path to the public key file for the server in ZeroMQ curve
            encryption.
        public_keys_location: str, optional
            The location to store the public keys for ZeroMQ curve encryption.
        secret_key: bytes, optional
            The secret key for ZeroMQ curve encryption.
        public_key: bytes, optional
            The public key for ZeroMQ curve encryption.
        server_public_key: bytes, optional
            The public key for the server in ZeroMQ curve encryption.

    Methods:
        _create_socket: zmq.Socket
            Creates and configures a ZeroMQ socket.
        _close_socket: None
            Closes the ZeroMQ socket and the context, and stops the authenticator
            if it was started.

    Examples:
        Create a REP socket and bind it to an address:

        >>> async with ZMQContextManager(zmq.REP, bind='tcp://*:5555') as socket:
        ...     while True:
        ...         message = await socket.recv()
        ...         await socket.send(message)

        Create a REQ socket and connect it to an address:

        >>> with ZMQContextManager(zmq.REQ, connect='tcp://localhost:5555') as socket:
        ...     socket.send(b'Hello')
        ...     message = socket.recv()
    """

    def __init__(self,
                 socket_type: zmq.SocketType,
                 bind: Optional[str] = None,
                 connect: Optional[str] = None,
                 secret_key_file: Optional[str] = None,
                 server_public_key_file: Optional[str] = None,
                 public_keys_location: Optional[str] = None,
                 secret_key: Optional[bytes] = None,
                 public_key: Optional[bytes] = None,
                 server_public_key: Optional[bytes] = None,
                 socket: Optional[zmq.Socket] = None):
        self.socket_type = socket_type
        if bind is None and connect is None:
            raise ValueError("Either 'bind' or 'connect' must be specified.")
        if bind is not None and connect is not None:
            raise ValueError("Both 'bind' and 'connect' cannot be specified.")
        self.bind = bind
        self.connect = connect
        self.secret_key = secret_key
        self.public_key = public_key
        self.server_public_key = server_public_key

        if secret_key_file:
            self.public_key, self.secret_key = zmq.auth.load_certificate(
                secret_key_file)

        if (self.secret_key is not None and self.public_key is None
                or self.secret_key is None and self.public_key is not None):
            raise ValueError(
                "Both secret key and public key must be specified.")

        if server_public_key_file:
            self.server_public_key = zmq.auth.load_certificate(
                server_public_key_file)[0]

        self.public_keys_location = public_keys_location

        self.observer = None
        self.auth = None
        self.context = None
        self.socket = None
        self._external_socket = socket

    def _create_socket(self, asyncio=False) -> zmq.Socket:
        """
        Creates and configures a ZeroMQ socket. Sets up security if required,
        and binds or connects the socket according to the specified settings.

        Returns:
            zmq.Socket: The configured ZeroMQ socket.
        """
        if self._external_socket:
            return self._external_socket
        if asyncio:
            self.context = zmq.asyncio.Context()
        else:
            self.context = zmq.Context.instance()

        self.socket = self.context.socket(self.socket_type)
        self.auth = None

        if self.bind and self.secret_key:
            if asyncio:
                self.auth = AsyncioAuthenticator(self.context)
            else:
                self.auth = ThreadAuthenticator(self.context)
            self.auth.start()
            self.reload_certificates()
            self.auto_reload_certificates()
            self.socket.curve_server = True  # must come before bind

        if self.secret_key:
            self.socket.curve_secretkey = self.secret_key
            self.socket.curve_publickey = self.public_key

        if self.bind:
            self.socket.bind(self.bind)
        if self.connect:
            if self.server_public_key:
                self.socket.curve_serverkey = self.server_public_key
            self.socket.connect(self.connect)
        return self.socket

    def reload_certificates(self):
        if self.public_keys_location and self.auth:
            self.auth.configure_curve(domain='*',
                                      location=self.public_keys_location)

    def auto_reload_certificates(self):
        self.observer = Observer()
        self.observer.schedule(ReloadCertificatesHandler(self),
                               self.public_keys_location,
                               recursive=False)
        self.observer.start()

    def _close_socket(self) -> None:
        """
        Closes the ZeroMQ socket and the context, and stops the authenticator
        if it was started.
        """
        if self._external_socket:
            return
        if self.observer:
            self.observer.stop()
            self.observer.join()
        self.socket.close()
        if self.auth:
            self.auth.stop()
        self.context.term()

    def __enter__(self) -> zmq.Socket:
        return self._create_socket(asyncio=False)

    def __exit__(self, exc_type: Optional[type], exc_val: Optional[Exception],
                 exc_tb: Optional[type]) -> None:
        self._close_socket()

    async def __aenter__(self) -> zmq.Socket:
        return self._create_socket(asyncio=True)

    async def __aexit__(self, exc_type: Optional[type],
                        exc_val: Optional[Exception],
                        exc_tb: Optional[type]) -> None:
        self._close_socket()



from .config import COL_SEL, COL_UNSEL, SymSize, defualt_colors, ridx, widths
from .qt_compat import QtWidgets

# the plotting widget
try:
    import pyqtgraph as pg
except:
    raise ImportError("Please install pyqtgraph first")

try:
    import pyperclip as pc
    hasCliper = True
except:
    hasCliper = False


class PlotWidget(pg.PlotWidget):

    def __init__(self, minimum_height=300, colors=None):
        self.XAxisLinked = False
        self.YAxisLinked = False
        if colors is None:
            colors = defualt_colors
        elif len(colors) < len(defualt_colors):
            colors.extend(defualt_colors[len(colors):])
        self.colors = colors
        self.xname = ""
        self.yname = ""
        super().__init__()

        self.setMinimumHeight(minimum_height)
        self.showGrid(x=True, y=True)
        self.setBackground(COL_UNSEL)

        self.plotItem.vb.autoRange()

        ## Labeling
        self.XLabel = QtWidgets.QLabel(self)
        self.XLabel.setText("X:")
        self.XLabel.move(0, 5)
        self.YLabel = QtWidgets.QLabel(self)
        self.YLabel.setText("Y:")
        self.YLabel.move(0, 35)

        self.plots = {}
        self.clippos1 = 0
        self.clippos2 = 0
        self.range_select = False
        for i in ridx:
            self.plots[i] = \
              self.plot([],[] ,pen={"color":self.colors[i]  ,"width":widths[i]} ,
                  symbolBrush = self.colors[i],
                  symbolPen = { "width":0 ,"color":self.colors[i] }   ,
                  symbolSize =SymSize[i] ,
              )
        self.update()

    def set_X_label(self, w):
        self.xname = w
        self.XLabel.setText(f"X:{w}")

    def set_Y_label(self, w):
        self.yname = w
        self.YLabel.setText(f"Y:{w}")

    def auto_range(self):
        self.plotItem.vb.autoRange()

    def enable_auto_range(self):
        self.plotItem.vb.enableAutoRange()

    def keyPressEvent(self, ev):
        #print(ev.text());
        tx = ev.text()
        if ('f' == tx or 'F' == tx):
            self.plotItem.vb.autoRange()
        if ('a' == tx or 'A' == tx):
            self.plotItem.vb.setAutoPan()
        if ('r' == tx or 'R' == tx):
            self.range_select = True
        super().keyPressEvent(ev)

    def keyReleaseEvent(self, ev):
        #print(ev.text());
        tx = ev.text()
        if ('f' == tx or 'F' == tx):
            self.plotItem.vb.autoRange()
        if ('a' == tx or 'A' == tx):
            self.plotItem.vb.setAutoPan()
        if ('r' == tx or 'R' == tx):
            self.range_select = False
        super().keyReleaseEvent(ev)

    def mousePressEvent(self, ev):
        if (4 == ev.button()):
            # print(ev.flags())
            if (hasCliper):
                # print("Mouse is pressed")
                self.clippos1 = self.plotItem.vb.mapSceneToView(ev.pos()).x()
        else:
            super().mousePressEvent(ev)

    def mouseReleaseEvent(self, ev):
        if (4 == ev.button()):
            p = ev.pos()
            if (hasCliper):
                # print("Mouse is released")
                self.clippos2 = self.plotItem.vb.mapSceneToView(ev.pos()).x()
                if (self.range_select):
                    pc.copy(f"{self.clippos1},{self.clippos2}")
                else:
                    pc.copy(self.clippos2)
        else:
            super().mouseReleaseEvent(ev)

    def update(self):
        super().update()

    def set_data(self, i, x, y):
        self.plots[i].setData(x, y)

    # def mouseDoubleClickEvent(self, ev):
    #     super().mouseDoubleClickEvent(ev)



from collections import defaultdict, deque

from .config import Nroll

PAUSE_TIME = 1

Number = int | float | complex


def remove_duplicates(input_list: list[str]) -> list[str]:
    """
    Remove duplicates from a list of strings, keeping the order of the elements.
    """
    return list(dict.fromkeys(input_list))


class Dataset():

    def __init__(self):
        self.column_names = []
        self.box = deque(maxlen=Nroll)
        self.dirty = True

    def clear(self):
        self.box.clear()

    def clear_history(self):
        o = self.box.popleft()
        self.box.clear()
        self.box.appendleft(o)

    def set_column_names(self, column_names: list[str]):
        column_names = remove_duplicates(column_names)
        if column_names != self.column_names:
            self.clear()
        self.column_names = column_names

    def get_data(self, step: int, xname: str,
                 yname: str) -> tuple[list[Number], list[Number]]:
        try:
            b = self.box[step]
        except IndexError:
            return [], []
        return b[xname], b[yname]

    def append(self, dataframe: list[Number] | list[list[Number]]):
        if not dataframe:
            return
        try:
            iter(dataframe[0])  # test if dataframe is a list of list
            self._append_traces(dataframe)
        except TypeError:
            self._append_points(dataframe)

    def roll(self):
        self.box.appendleft(defaultdict(list))

    def _append_points(self, points: list[Number]):
        self.dirty = True
        assert (len(points) == len(
            self.column_names)), (f"-PointDataBox\n"
                                  f"-ap\n"
                                  f"-Length Must be same\n"
                                  f"the column_names : {self.column_names}\n"
                                  f"given data : {points}")
        for name, p in zip(self.column_names, points):
            self.box[0][name].append(p)

    def _append_traces(self, traces: list[list[Number]]):
        self.dirty = True
        assert (len(traces) == len(
            self.column_names)), (f"-TraceDataBox\n"
                                  f"-at\n"
                                  f"-Length Must be same\n"
                                  f"the column_names : {self.column_names}\n"
                                  f"given data : {traces}")
        self.box.appendleft(dict(zip(self.column_names, traces)))



import functools
import operator

import matplotlib.pyplot as plt
import numpy as np
from matplotlib import cm
from matplotlib.colors import Normalize

layout_example = {
    'qubits': {
        'Q0': {
            'pos': (0, 1)
        },
        'Q1': {
            'pos': (1, 0)
        },
        'Q2': {
            'pos': (0, -1)
        },
        'Q3': {
            'pos': (-1, 0)
        }
    },
    'couplers': {
        'C0': {
            'qubits': ['Q0', 'Q1'],
        },
        'C1': {
            'qubits': ['Q1', 'Q2'],
        },
        'C2': {
            'qubits': ['Q2', 'Q3'],
        },
        'C3': {
            'qubits': ['Q0', 'Q3'],
        }
    }
}


def complete_layout(layout):
    for c in layout['couplers']:
        qubits = layout['couplers'][c]['qubits']
        for q in qubits:
            if q not in layout['qubits']:
                raise ValueError('qubit {} not found'.format(q))
            if 'couplers' not in layout['qubits'][q]:
                layout['qubits'][q]['couplers'] = []
            if c not in layout['qubits'][q]['couplers']:
                layout['qubits'][q]['couplers'].append(c)
    return layout


def get_shared_coupler(layout, q1, q2):
    for c in layout['qubits'][q1]['couplers']:
        if q2 in layout['couplers'][c]['qubits']:
            return c
    return None


def get_neighbours(layout,
                   qubit_or_coupler,
                   distance=1,
                   type='qubit',
                   inrange=False):

    def _qubits(couplers):
        ret = set()
        for c in couplers:
            ret = ret | set(layout['couplers'][c]['qubits'])
        return ret

    def _couplers(qubits):
        ret = set()
        for q in qubits:
            ret = ret | set(layout['qubits'][q]['couplers'])
        return ret

    couplers = []
    neighbors = []

    if qubit_or_coupler in layout['qubits']:
        couplers.append(set(layout['qubits'][qubit_or_coupler]['couplers']))
        neighbors.append(_qubits(couplers[0]) - {qubit_or_coupler})
    elif qubit_or_coupler in layout['couplers']:
        if type == 'coupler':
            distance += 1
        neighbors.append(set(layout['couplers'][qubit_or_coupler]['qubits']))
        couplers.append({qubit_or_coupler})
    else:
        raise ValueError(f'qubit or coupler {qubit_or_coupler!r} not found')
    distance -= 1

    while distance > 0:
        couplers.append(_couplers(neighbors[-1]) - couplers[-1])
        neighbors.append(_qubits(couplers[-1]) - neighbors[-1])
        distance -= 1

    if type == 'qubit':
        if inrange:
            return list(functools.reduce(operator.or_, neighbors, set()))
        else:
            return list(neighbors[-1])
    elif type == 'coupler':
        if inrange:
            if qubit_or_coupler in couplers[0]:
                couplers = couplers[1:]
            return list(functools.reduce(operator.or_, couplers, set()))
        else:
            return list(couplers[-1])
    else:
        raise ValueError("type must be 'qubit' or 'coupler'")


def plot_range(ax,
               path,
               text='',
               color=None,
               text_color='k',
               bounder_color='k',
               lw=0.5,
               fontsize=9):
    x, y = path
    center = x.mean(), y.mean()

    if color:
        ax.fill(x, y, color=color, lw=0)

    if lw is not None and lw > 0:
        ax.plot(np.hstack([x, [x[0]]]),
                np.hstack([y, [y[0]]]),
                color=bounder_color,
                lw=lw)

    if text:
        ax.text(center[0],
                center[1],
                text,
                ha='center',
                va='center',
                color=text_color,
                fontsize=fontsize)


def circle_path(pos, r, n=40):
    x, y = pos
    t = 2 * np.pi * np.linspace(0, 1, n, endpoint=False)
    xx = r * np.cos(t) + x
    yy = r * np.sin(t) + y
    return xx, yy


def circle_link_path(pos1, pos2, r1, r2, width, n=20):
    width = min(2 * max(r1, r2), width)

    x1, y1 = pos1
    x2, y2 = pos2

    phi = np.arctan2(y2 - y1, x2 - x1)

    theta1 = np.arcsin(width / 2 / r1)
    theta2 = np.arcsin(width / 2 / r2)

    t = np.linspace(-theta1, theta1, n) + phi
    xx1 = r1 * np.cos(t) + x1
    yy1 = r1 * np.sin(t) + y1

    t = np.linspace(-theta2, theta2, n) + phi + np.pi
    xx2 = r2 * np.cos(t) + x2
    yy2 = r2 * np.sin(t) + y2

    return np.hstack([xx2[-1], xx1,
                      xx2[:-1]]), np.hstack([yy2[-1], yy1, yy2[:-1]])


def circle_half_directed_link_path(pos1, pos2, r1, r2, width, n=20):
    width = min(max(r1, r2), width)

    x1, y1 = pos1
    x2, y2 = pos2

    phi = np.arctan2(y2 - y1, x2 - x1)

    theta1 = np.arcsin(width / r1)
    theta2 = np.arcsin(width / r2)

    t = np.linspace(0.2 * theta1, theta1, n) + phi
    xx1 = r1 * np.cos(t) + x1
    yy1 = r1 * np.sin(t) + y1

    t = np.linspace(-theta2, -0.2 * theta2, n) + phi + np.pi
    xx2 = r2 * np.cos(t) + x2
    yy2 = r2 * np.sin(t) + y2

    v = (xx2[0] - xx1[-1]) + 1j * (yy2[0] - yy1[-1])
    c = (xx2[0] + xx1[-1]) / 2 + 1j * (yy2[0] + yy1[-1]) / 2

    a = np.array([1 / 6, 1 / 12]) + 1j * np.array([0, 0.4 * width / np.abs(v)])
    a = a * v + c

    return np.hstack([xx2[-1], xx1, a.real,
                      xx2[:-1]]), np.hstack([yy2[-1], yy1, a.imag, yy2[:-1]])


def draw(layout, ax=None, qubit_cbar=True, coupler_cbar=True, origin='upper'):
    if ax is None:
        ax = plt.gca()

    for qubit in layout['qubits'].values():
        pos = qubit['pos']
        if origin == 'upper':
            pos = pos[0], -pos[1]
        path = circle_path(pos, qubit.get('radius', 0.5))
        plot_range(ax,
                   path,
                   qubit.get('text', ''),
                   qubit.get('color', None),
                   lw=qubit.get('lw', 0.5),
                   fontsize=qubit.get('fontsize', 9),
                   text_color=qubit.get('text_color', 'k'),
                   bounder_color=qubit.get('bounder_color', 'k'))

    for coupler in layout['couplers'].values():
        q1, q2 = coupler['qubits']
        pos1 = layout['qubits'][q1]['pos']
        pos2 = layout['qubits'][q2]['pos']
        if origin == 'upper':
            pos1 = pos1[0], -pos1[1]
            pos2 = pos2[0], -pos2[1]
        r1 = layout['qubits'][q1].get('radius', 0.5)
        r2 = layout['qubits'][q2].get('radius', 0.5)
        width = coupler.get('width', 0.5)
        lw = coupler.get('lw', 0.5)

        path = circle_link_path(pos1, pos2, r1, r2, width)
        plot_range(ax,
                   path,
                   coupler.get('text', ''),
                   color=coupler.get('color', None),
                   lw=0,
                   fontsize=coupler.get('fontsize', 9),
                   text_color=coupler.get('text_color', 'k'))
        if lw > 0:
            x, y = circle_link_path(pos1, pos2, r1, r2, width, n=2)
            ax.plot(x[:2],
                    y[:2],
                    lw=lw,
                    color=coupler.get('bounder_color', 'k'))
            ax.plot(x[2:],
                    y[2:],
                    lw=lw,
                    color=coupler.get('bounder_color', 'k'))

    ax.axis('equal')
    ax.set_axis_off()

    if qubit_cbar and layout['__colorbar__']['qubit']['norm'] is not None:
        cbar = plt.colorbar(cm.ScalarMappable(
            norm=layout['__colorbar__']['qubit']['norm'],
            cmap=layout['__colorbar__']['qubit']['cmap']),
                            ax=ax,
                            location='bottom',
                            orientation='horizontal',
                            pad=0.01,
                            shrink=0.5)
        cbar.set_label(layout['__colorbar__']['qubit']['label'])
    if coupler_cbar and layout['__colorbar__']['coupler']['norm'] is not None:
        cbar = plt.colorbar(cm.ScalarMappable(
            norm=layout['__colorbar__']['coupler']['norm'],
            cmap=layout['__colorbar__']['coupler']['cmap']),
                            ax=ax,
                            location='bottom',
                            orientation='horizontal',
                            pad=0.01,
                            shrink=0.5)
        cbar.set_label(layout['__colorbar__']['coupler']['label'])


def get_norm(params, elms, vmin=None, vmax=None):
    data = []
    for elm in elms:
        if elm in params:
            if isinstance(params[elm], (int, float)):
                data.append(params[elm])
            elif 'value' in params[elm] and params[elm]['value'] is not None:
                data.append(params[elm]['value'])
    if data:
        if vmin is None:
            vmin = min(data)
        if vmax is None:
            vmax = max(data)
        return Normalize(vmin=vmin, vmax=vmax)
    else:
        return None


def fill_layout(layout,
                params,
                qubit_size=0.5,
                coupler_size=0.5,
                qubit_fontsize=9,
                coupler_fontsize=9,
                qubit_color=None,
                coupler_color=None,
                qubit_cmap='hot',
                qubit_vmax=None,
                qubit_vmin=None,
                qubit_norm=None,
                coupler_cmap='binary',
                coupler_vmax=None,
                coupler_vmin=None,
                coupler_norm=None,
                bounder_color='k',
                lw=0.5):

    qubit_cmap = plt.get_cmap(qubit_cmap)
    coupler_cmap = plt.get_cmap(coupler_cmap)

    if qubit_norm is None:
        qubit_norm = get_norm(params,
                              layout['qubits'].keys(),
                              vmin=qubit_vmin,
                              vmax=qubit_vmax)
    if coupler_norm is None:
        coupler_norm = get_norm(params,
                                layout['couplers'].keys(),
                                vmin=coupler_vmin,
                                vmax=coupler_vmax)
    layout['__colorbar__'] = {
        'coupler': {
            'cmap': coupler_cmap,
            'norm': coupler_norm,
            'label': ''
        },
        'qubit': {
            'cmap': qubit_cmap,
            'norm': qubit_norm,
            'label': ''
        }
    }

    for qubit in layout['qubits']:
        layout['qubits'][qubit]['radius'] = qubit_size
        layout['qubits'][qubit]['fontsize'] = qubit_fontsize
        if qubit in params:
            layout['qubits'][qubit]['lw'] = 0
            if not isinstance(params[qubit], dict):
                params[qubit] = {'value': params[qubit]}
            if 'color' in params[qubit]:
                layout['qubits'][qubit]['color'] = params[qubit]['color']
            elif 'value' in params[qubit] and params[qubit][
                    'value'] is not None:
                layout['qubits'][qubit]['color'] = qubit_cmap(
                    qubit_norm(params[qubit]['value']))
            else:
                layout['qubits'][qubit]['color'] = qubit_color
                if qubit_color is None:
                    layout['qubits'][qubit]['lw'] = lw
            layout['qubits'][qubit]['radius'] = params[qubit].get(
                'radius', qubit_size)
            layout['qubits'][qubit]['fontsize'] = params[qubit].get(
                'fontsize', qubit_fontsize)
            layout['qubits'][qubit]['text'] = params[qubit].get('text', '')
            layout['qubits'][qubit]['text_color'] = params[qubit].get(
                'text_color', 'k')
        else:
            layout['qubits'][qubit]['color'] = qubit_color
            if qubit_color is None:
                layout['qubits'][qubit]['lw'] = lw
            else:
                layout['qubits'][qubit]['lw'] = 0
            layout['qubits'][qubit]['bounder_color'] = bounder_color

    for coupler in layout['couplers']:
        layout['couplers'][coupler]['width'] = coupler_size
        layout['couplers'][coupler]['fontsize'] = coupler_fontsize
        layout['couplers'][coupler]['bounder_color'] = bounder_color
        if coupler in params:
            layout['couplers'][coupler]['lw'] = 0
            if not isinstance(params[coupler], dict):
                params[coupler] = {'value': params[coupler]}
            if 'color' in params[coupler]:
                layout['couplers'][coupler]['color'] = params[coupler]['color']
            elif 'value' in params[coupler] and params[coupler][
                    'value'] is not None:
                layout['couplers'][coupler]['color'] = coupler_cmap(
                    coupler_norm(params[coupler]['value']))
            else:
                layout['couplers'][coupler]['color'] = coupler_color
                if coupler_color is None:
                    layout['couplers'][qubit]['lw'] = lw
            layout['couplers'][coupler]['width'] = params[coupler].get(
                'width', coupler_size)
            layout['couplers'][coupler]['fontsize'] = params[coupler].get(
                'fontsize', coupler_fontsize)
            layout['couplers'][coupler]['text'] = params[coupler].get(
                'text', '')
            layout['couplers'][coupler]['text_color'] = params[coupler].get(
                'text_color', 'k')
        else:
            layout['couplers'][coupler]['color'] = coupler_color
            if coupler_color is None:
                layout['couplers'][coupler]['lw'] = lw
            else:
                layout['couplers'][coupler]['lw'] = 0
            layout['couplers'][coupler]['bounder_color'] = bounder_color

    return layout
